# keysig.py

"""
The program "MMA - Musical Midi Accompaniment" and the associated
modules distributed with it are protected by copyright.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Keysignature is used in the solo/melody tracks and for midinote.
A single instance class is created here and can be accessed from any
other module.

"""

from . import gbl
from   mma.MMA.common import *
import mma.MMA.debug

majKy = { "C" :  0, "G" :  1, "D" :  2,
          "A" :  3, "E" :  4, "B" :  5,
          "F#":  6, "C#":  7, "F" : -1,
          "Bb": -2, "Eb": -3, "Ab": -4,
          "Db": -5, "Gb": -6, "Cb": -7 }

minKy = { "A" :  0, "E" :  1, "B" :  2,
          "F#":  3, "C#":  4, "G#":  5,
          "D#":  6, "A#":  7, "D" : -1,
          "G" : -2, "C" : -3, "F" : -4,
          "Bb": -5, "Eb": -6, "Ab": -7 }

notevalues={
        'B#': 0, 'C' : 0, 'C#': 1, 'Db': 1,
        'D' : 2, 'D#': 3, 'Eb': 3, 'E' : 4,
        'Fb': 4, 'E#': 5, 'F' : 5, 'F#': 6,
        'Gb': 6, 'G' : 7, 'G#': 8, 'Ab': 8,
        'A' : 9, 'A#': 10,'Bb': 10,'B' : 11,
        'Cb':11 }


class KeySig:

    def __init__(self):
        self.kSig = 0
        self.kName = ['C', 0]
        self.setAccList()
        self.keyNoteValue = notevalues['C']

    def create(self, ln):
        """ Set the keysignature. Used by solo & aria tracks. Formats are:
              1. A,D,E, Eb,  etc followed by optional minor/major
                   flat can be b or &, sharp is #
              2. Number of sharps/flats plus option minor/major
                   id 3b or 2#
        """

        mi = 0   # assume Major

        if len(ln) < 1 or len(ln) > 2:
            error("KeySig: Needs 1 or 2 arguments.")

        # got 2 args, parse off minor/major

        if len(ln) == 2:
            l = ln[1][0:3].upper()
            if l == 'MIN':
                mi = 1
            elif l == 'MAJ':
                mi = 0
            else:
                error("KeySig: 2nd arg must be 'Major' or 'Minor', not '%s'" % ln[1])

        self.kName[1] = mi  # remember maj/min setting

        # get the key

        kname = ln[0][0]
        kname = kname.upper()
        if len(ln[0]) > 1:
            kname += ln[0][1:]

        if kname[0] in "ABCDEFG":
            if len(kname) == 2 and kname[1] == '&':
                kname = kname[0] + 'b'
            if mi and kname in minKy:
                self.kSig = minKy[kname]
                midikey = minKy[kname]
            elif not mi and kname in majKy:
                self.kSig = majKy[kname]
                midikey = majKy[kname]
            else:
                if mi:
                    t = 'minor'
                else:
                    t = 'major'
                error("KeySigs: keysignature '%s %s' is unknown/impossible." % (kname, t))
            self.kName[0] = kname  # save name ('C', "Eb", etc)

        elif kname[0] in "01234567":
            c = int(kname[0])

            if kname == '0':
                kname = '0#'
                
            if len(kname) < 2:
                error("KeySig: numerical keysig needs sharp/flat indicator.")

            f = kname[1].upper()

            if not f in ("B", "&", "#"):
                error("KeySig: 2nd char in KeySig must be 'b' or '#', not '%s'" % f)

            self.kSig = midikey = int(c)

            if f in ('B', '&'):
                self.kSig = -self.kSig

            if mi:
                z = minKy
            else:
                z = majKy
            for a in z:
                if z[a] == self.kSig:
                    self.kName[0] = a
                    break

        else:
            error("KeySig: unknown keysignature '%s'." % ln[0])

        # Set the midi meta track with the keysig. This doen't do anything
        # in the playback, but other programs may use it.

        if midikey < 0:
            midikey = 256 + midikey

        gbl.mtrks[0].addKeySig(gbl.tickOffset, midikey, mi)

        self.setAccList()

        self.keyNoteValue = notevalues[self.kName[0]]

        # Check for aria tracks. If found, reset the note list to null.
        for t in gbl.tnames:
            if gbl.tnames[t].vtype == 'ARIA':
                gbl.tnames[t].restart()

        if mma.MMA.debug.debug:
            dPrint("KeySig: %s" % self.getKeysig())

    def getKeysig(self):
        """ Create a key sig string. """

        return "%s %s" % (self.kName[0], ('Major', 'Minor')[self.kName[1]])

    def setAccList(self):
        """ Create a keysig table. This table is created in __init__ and
            when a keysig changes. There is an entry for each note,
            either -1,0,1 corresponding to flat,natural,sharp. We populate
            the table for each bar from the keysig value.

            Melody/solo tracks are given a copy of the table when they can
            then modify.
         """

        acc = {'a': 0, 'b': 0, 'c': 0, 'd': 0, 'e': 0, 'f': 0, 'g': 0}

        # Current key negative values are flat, positive sharp
        ks = self.kSig

        if ks < 0:
            for a in range(abs(ks)):
                acc[['b', 'e', 'a', 'd', 'g', 'c', 'f'][a]] = -1

        else:
            for a in range(ks):
                acc[['f', 'c', 'g', 'd', 'a', 'e', 'b'][a]] = 1

        self.accList = acc

keySig = KeySig()    # single instance


def getTranspose(ln, err):
    """ Calc transpose. Used by global TRANSPOSE and Lyric Transpose= """

    # interval notation accepts Up-Perfect-Four (with hypens)
    if len(ln) == 1 and ln[0].count('-') == 2:
        ln = ln[0].split('-')
        
    # Simple semi-tone transpose. Just make sure it's 0 to 12
    if len(ln) == 1:
        value = stoi(ln[0], "%s: Argument must be an integer, not '%s'" % (err, ln[0]))
        if value < -12 or value > 12:
            error("%s: %s out-of-range; must be -12..12" % (err, value))

    # interval transposition uses "up/down interval
    # this needs exactly 3 args. eg: "Up Perfect 3"
    # Note: This does NOT properly set key name spellings (ie: C# vrs Db)
    elif len(ln) == 3:

        
        # see if the input is a unique abrev. At some point we should put
        # this in common.py ... if we decide to do more abrev. matching.
        def getAbrev(i, tb):
            mcount = 0
            i = i.upper()
            for f in tb:
                if f.startswith(i):
                    mcount += 1
                    match = f
            if mcount > 1:
                error("%s: '%s' is not a unique abreviation." % (err, i))
            if mcount < 1:
                error("%s: '%s' is unregonized for interval notation." % (err,i))
            return match

        num2text = {"1": "UNISON",
                    "2": "SECOND",
                    "3": "THIRD",
                    "4": "FOURTH",
                    "5": "FIFTH",
                    "6": "SIXTH",
                    "7": "SEVENTH",
                    "8": "OCTAVE"}
        
        ttable = {"PERFECT UNISON":     0,
                  "DIMINISHED SECOND":  0,
                  "AUGMENTED UNISON":   1,
                  "MINOR SECOND":       1,
                  "MAJOR SECOND":       2,
                  "DIMINISHED THIRD":   2,
                  "AUGMENTED SECOND":   3,
                  "MINOR THIRD":        3,
                  "MAJOR THIRD":        4,
                  "DIMINISHED FOURTH":  4,
                  "AUGMENTED THIRD":    5,
                  "PERFECT FOURTH":     5,
                  "AUGMENTED FOURTH":   6,
                  "DIMINISHED FIFTH":   6,
                  "PERFECT FIFTH":      7,
                  "DIMINISHED SIXTH":   7,
                  "AUGMENTED FIFTH":    8,
                  "MINOR SIXTH":        8,
                  "MAJOR SIXTH":        9,
                  "DIMINISHED SEVENTH": 9,
                  "AUGMENTED SIXTH":   10,
                  "MINOR SEVENTH":     10,
                  "MAJOR SEVENTH":     11,
                  "DIMINISHED OCTAVE": 11,
                  "PERFECT OCTAVE":    12 }

        # Each token might be an abrev. Convert them to full spellings.

        # 1st, get direction
        dir = getAbrev(ln[0], ('UP', 'DOWN'))
        if dir == 'UP':
            mul = 1
        elif dir == 'DOWN':
            mul = -1
        else:
            error("%s: Interval transpose needs 'UP/DOWN' as first arg, not '%s'." % (err, dir))

        # Get the interval

        if ln[2] in num2text:
            ln[2] = num2text[ln[2]]
            
        interval = '%s %s' % (getAbrev(ln[1], set([ a.split()[0] for a in ttable.keys()])),
                getAbrev(ln[2], set([ a.split()[1] for a in ttable.keys()])))
       
        if interval in ttable:
            value = ttable[interval]
        else:
            error("%s: Unknown/invalid interval '%s'. Valid keys are: %s." % \
                  (err, interval, ', '.join(ttable.keys())))

        value *= mul
        
    else:
        error("%s: Use single value for semitones (-3, 4) or proper "
              "interval ('Up Maj 3'), not '%s'." % (err, ' '.join(ln)))

    return value

def transpose(ln):
    """ Set transpose value. """

    if len(ln) and ln[0].upper() == 'ADD':
        ln = ln[1:]
        gbl.transpose += getTranspose(ln, "Transpose")
        if gbl.transpose < -12 or gbl.transpose > 12:
            error("Transose: Add option results in setting of '%s', must be -12 to 12." %
                  gbl.transpose)
    else:
        gbl.transpose = getTranspose(ln, "Transpose")

    if mma.MMA.debug.debug:
        dPrint("Set Transpose to %s" % gbl.transpose)



# timesig.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Timesig manager. All timesig stuff is here, except for the Mtrk.setTimeSig
class function. For a function which has no effect, it appears to be
a lot of code. But some sequencers and notation programs absolutely require
proper time sigs for correct operation. So, here you go.
"""

from . import gbl
import mma.MMA.midi
from   mma.MMA.common import *

class TimeSig:
    """ Track and set the current time signature.

        Timesigs are completely optional and are inserted into
        the MIDI file by addTimeSig(). MMA routines ignore timesig
        settings.

        NOTE: The actual func to place a timesig meta into the midi 
              stream is not here. It's part of the Mtrk() class.
    """

    def __init__(self):
        """ Initialze to null value, user will never set to this."""

        self.lastsig = (None, None)


    def create(self, nn, dd):
        """ Set timesig. If no change from last value, ignore. """

        if self.lastsig != (nn, dd):
            gbl.mtrks[0].addTimeSig(gbl.tickOffset, nn, dd, 48, 8)
            self.lastsig = (nn, dd)


    def setSig(self, ln):
        """ Set the midi time signature from parser. """

        if len(ln) == 1:
            a = ln[0].upper()
            if a == 'COMMON':
                ln = ('4', '4')
            elif a == 'CUT':
                ln = ('2', '2')
            elif '/' in ln[0]:
                ln = ln[0].split('/',1)

        if len(ln) != 2:
            error("TimeSig: Usage (numerator denominator) or ('cut' or 'common')")

        nn = stoi(ln[0])

        if nn < 1 or nn > 126:
            error("Timesig: Numerator must be 1..126")

        denominators = {'1':0, '2':1, '4':2, '8':3, '16':4, '32':5, '64':6 }
        dd = ln[1]

        if dd in denominators:
            dd = denominators[dd]

        else:
            error("Timesig: Denominator must be 1, 2, 4, 8, 16, 32 or 64, not '%s'." % dd)

        self.create(nn, dd)


    def get(self):
        """ Return existing timesig in MIDI format. """

        return self.lastsig


    def getAscii(self):
        """ Return existing timesig in readable format. """
        
        if not self.lastsig[0]:
            return "Not Set"
        n, d = self.lastsig
        return "%s/%s" % (n, ['1', '2', '4', '8',  '16', '32', '64'][d] )


timeSig = TimeSig()  # singleton 





# paths.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module contains functions for setting various path variables.

"""

import os

from . import gbl
from mma.MMA.common import *
import mma.MMA.auto
import mma.MMA.grooves
import mma.MMA.exits
import mma.MMA.debug

outfile = ''

libPath = []
libDirs = []
incPath = []
plugPaths = []

mmaStart = []
mmaEnd   = []
mmaRC    = None

def init():
    """ Called from main. In mma.MMA.py we checked for known directories and
        inserted the first found 'mma' directory into the sys.path list and
        set MMAdir. Now, set the lib/inc lists.
    """
    
    setLibPath([os.path.join(gbl.MMAdir, 'lib')], user=0)
    if not libPath or not os.path.isdir(libPath[0]):
        dPrint("Warning: Library directory not found (check mma.MMA.py).")

    setIncPath([os.path.join(gbl.MMAdir, 'includes')])
    if not incPath or not os.path.isdir(incPath[0]):
        dPrint("Warning: Include directory not found.")

    setPlugPath([os.path.join(gbl.MMAdir, "plugins")] )


##################################
# Set up the mma start/end paths

def mmastart(ln):
    """ Set/append to the mmastart list. """

    if not ln:
        error("Use: MMAstart FILE [file...]")

    for a in ln:
        gbl.mmaStart.append(mma.MMA.file.fixfname(a))

    if mma.MMA.debug.debug:
        dPrint("MMAstart set to: %s" % gbl.mmaStart)

def mmaend(ln):
    """ Set/append to the mmaend list. """

    if not ln:
        error("Use: MMAend FILE [file...]")

    for a in ln:
        gbl.mmaEnd.append(mma.MMA.file.fixfname(a))

    if mma.MMA.debug.debug:
        dPrint("MMAend set to: %s" % gbl.mmaEnd)

def setRC(f):
    """ Set a rc file from the command line."""
    
    global mmaRC

    mmaRC = f

######################################
# process the RC, mmastart and mmaend files. Called from main.py

def readRC():
    """ Process all RC files. """

    docOption = gbl.createDocs   # Disable doc printing for RC file
    gbl.createDocs = 0

    if mmaRC:
        rcfiles = [mmaRC]
    else:
        rcfiles = ('mmarc', 'c:\\mma\\mmarc', '~/.mmarc', '/usr/local/etc/mmarc', '/etc/mmarc')

    readDone = 0
    for i in rcfiles:
        f = mma.MMA.file.locFile(i, None)
        if f:
            if mma.MMA.debug.showrun:
                dPrint("Reading RC file '%s'" % f)
            mma.MMA.parse.parseFile(f)
            readDone = 1
            break
        else:
            if mmaRC:
                error("Specified init file '%s' not found" % mmaRC)

    if not readDone and mma.MMA.debug.debug:
        gbl.lineno = -1
        warning("No RC file was found or processed")

    gbl.createDocs = docOption   # Restore doc options


def dommaStart():
    """ Process all the mma start files. """

    for f in mmaStart:
        fn = findIncFile(f)
        if not fn:
            warning("MmaStart file '%s' not found/processed" % f)
        else:
            mma.MMA.parse.parseFile(fn)
        gbl.lineno = -1   # reset for real code


def dommaEnd():
    """ Process all the mma end files."""

    for f in mmaStart:
        fn = findIncFile(f)
        if not fn:
            warning("MmaStart file '%s' not found/processed" % f)
        else:
            mma.MMA.parse.parseFile(fn)
        gbl.lineno = -1   # reset for real code


#######################################
# Search the paths for a file.


def findIncFile(fn):
    """ Find an INC file. Returns complete path or NULL."""

    global incPath

    for lib in incPath:
        path = mma.MMA.file.locFile(fn, lib)
        if path:
            return path

    return None


def findLibFile(fn):
    """ Find a LIB file. Returns complete path or NULL."""

    global libDirs

    if not libDirs:
        expandLib()

    for lib in libDirs:
        path = mma.MMA.file.locFile(fn, lib)
        if path:
            return path

    return None


##############################################
# Set up the lib/inc paths

def setLibPath(ln, user=1):
    """ Set the LibPath variable.  """

    global libPath, libDirs
    libPath = []
    libDirs = []

    for l in ln:
        f = mma.MMA.file.fixfname(l)
        libPath.append(f)

    expandLib(user)

    if mma.MMA.debug.debug:
        dPrint("LibPath set: %s" % ' '.join(libPath))


def expandLib(user=0):
    """ Expand the library paths from the list in libdir. """

    global libPath, libDirs

    scount = 0

    # Parse all the lib trees. Root trees are included into our list

    libDirs = []
    for f in libPath:
        for root, dir, files in os.walk(f):
            if root not in libDirs:
                if os.path.basename(root) == 'stdlib':
                    scount += 1
                    if not user:  # system init, stdlib goes first
                        libDirs.insert(0, root)
                    else:
                        libDirs.append(root)
                    continue
                libDirs.append(root)

    if not scount and not user:
        warning("Your library set does not have a 'stdlib'.")

    # forget about previously loaded mma lib databases

    mma.MMA.auto.grooveDB = []

    if mma.MMA.debug.debug:
        dPrint("LibPath expansion set to: %s" % ' '.join(libDirs))


def setIncPath(ln):
    """ Set the IncPath variable.  """

    global incPath
    incPath = []

    for l in ln:
        f = mma.MMA.file.fixfname(l)
        incPath.append(f)

    if mma.MMA.debug.debug:
        dPrint("IncPath set: %s" % ' '.join(incPath))

###########################################
# Output pathname


def setOutPath(ln):
    """ Set the Outpath variable. """

    if not ln:
        gbl.outPath = ""

    elif len(ln) > 1:
        error("Use: SetOutPath PATH")

    else:
        gbl.outPath = mma.MMA.file.fixfname(ln[0])

    if mma.MMA.debug.debug:
        dPrint("OutPath set to '%s'" % gbl.outPath)


def createOutfileName(extension):
    """ Create the output filename.

       Called from the mainline, below and from lyrics karmode.

       If outfile was specified on cmd line then leave it alone.
       Otherwise ...
         1. strip off the extension if it is .mma,
         2. append .mid
   """

    global outfile

    if gbl.playFile and gbl.outfile:
        error("You cannot use the -f option with -P")

    if gbl.outfile:
        outfile = gbl.outfile

    elif gbl.playFile:
        outfile = "MMAtmp%s.mid" % os.getpid()
        mma.MMA.exits.files.append(outfile)

    else:
        outfile, ext = os.path.splitext(gbl.infile)
        if ext != gbl.EXT:
            outfile = gbl.infile
        outfile += extension

    outfile = mma.MMA.file.fixfname(outfile)


##############################################
# Set up the plugin paths

def setPlugPath(ln):
    """ Set the plugPath variable.  """

    global plugPaths
    plugPaths = []

    for l in ln:
        plugPaths.append( mma.MMA.file.fixfname(l) )

    if mma.MMA.debug.debug:
        dPrint("PlugPath set: %s" % ' '.join(plugPaths))



# options.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import getopt
import sys
import os

import mma.MMA.docs
import mma.MMA.parse
import mma.MMA.chords
import mma.MMA.volume
import mma.MMA.exits

from . import gbl
from mma.MMA.common import *
from mma.MMA.macro import macros

cmdSMF = None

def cmdLine(l):
    """ Try to take a mma input line and parse it like a command line. 
        An empty CmdLine is just ignored. """
    
    if l:
        opts(l)

def cmdError(e):
    """ Illegal internal command line options."""

    error("CmdLine: the command line option '%s' is not permitted in a MMA script." % e)

def opts(l=None):
    """ Option parser. 
         FIXME: this code segment is much too long!
    """

    if not l:
        l = sys.argv[1:]
        internal = False
    else:
        internal = True

    try:
        opts, args = getopt.gnu_getopt(
                         l, "b:B:dpsS:ri:wneom:f:M:cLgGvVD:01PT:I:x:", [])
    except getopt.GetoptError:
        usage()

    for o, a in opts:
        if o == '-b':
            setBarRange(a)

        elif o == '-B':
            setBarRange(a)
            gbl.barRange.append("ABS")

        elif o in ('-d', '-o', '-p', '-s', '-r', '-w', '-n', '-e', '-c'):
            import mma.MMA.debug   # circular dep. problem
            mma.MMA.debug.cmdLineDebug(o[-1])
        
        elif o == '-S':
            ln = a.split('=', 1)
            macros.setvar(ln)

        elif o == '-L':
            gbl.printProcessed = True

        elif o == '-f':
            import mma.MMA.paths
            gbl.outfile = a
            if internal:
                warning("Output filename overwritten by -f CmdLine option.")
                mma.MMA.paths.createOutfileName(".mid")

        elif o == '-i':
            import mma.MMA.paths
            if internal:
                cmdError("-i")
            mma.MMA.paths.setRC(a)

        elif o == '-g':
            if internal:
                cmdError("-g")
            gbl.makeGrvDefs = 1

        elif o == '-G':
            if internal:
                cmdError("-G")
            gbl.makeGrvDefs = 2

        elif o == '-m':
            try:
                a = int(a)
            except:
                error("Expecting -m arg to be a integer")
            gbl.maxBars = a

        elif o == '-v':
            print("%s" % gbl.version)
            if not internal:
                sys.exit(0)

        elif o == '-M':
            global cmdSMF
            if a in ['0', '1']:
                cmdSMF = a
            else:
                error("Only a '0' or '1' is permitted for the -M arg")

        elif o == '-T':   # set tracks to generate, mute all others
            gbl.muteTracks = a.upper().split(',')

        elif o == '-D':
            if internal:
                cmdError("-D..")
            if a == 'xl':
                gbl.createDocs = 1

            elif a == 'xh':
                gbl.createDocs = 2

            elif a == 's':
                gbl.createDocs = 3

            elif a == 'gh':
                gbl.createDocs = 4

            elif a == 'bo':
                gbl.createDocs = 99

            elif a == 'k':
                import mma.MMA.alloc
                # important! Needs a space before the trailing LF for mma.MMA.el
                print("Base track names: %s \n" % 
                      ' '.join([a for a in sorted(mma.MMA.alloc.trkClasses)]))
                print("Commands: %s BEGIN END DEFAULT\n" % 
                      ' '.join([a for a in sorted(mma.MMA.parse.simpleFuncs)]))
                print("TrackCommands: %s \n" %
                      ' '.join([a for a in sorted(mma.MMA.parse.trackFuncs)]))
                print("Not complete ... subcommands, comments, chords...")
                sys.exit(0)

            else:
                print("Unknown option: '-D%s'." % a)
                usage()

        elif o == '-0':
            import mma.MMA.sync
            mma.MMA.sync.synchronize(['START'])

        elif o == '-1':
            import mma.MMA.sync
            mma.MMA.sync.synchronize(['END'])

        elif o == '-P':
            gbl.playFile = 1

        elif o == '-I':
            # We use -I for plugin help and overload it to discard 
            # the plugin security. Use -II for security override.
            # It does mean you can't have plugin called "I", but
            # you could use "i" and it'll work.
            import mma.MMA.regplug
            if a == 'I':
                mma.MMA.regplug.secOverRide = True

            # Plugin help. Note we have not loaded any plugins at this
            # point. pluginHelp() will find the plugin, register it and
            # call its help function.
            else: 
                mma.MMA.regplug.pluginHelp(a)
                sys.exit(0)
                

        elif o == '-V':
            import mma.MMA.file
            if internal:
                cmdError("-V")
            gbl.playFile = 2  # signal create and play groove
            if not args:
                error("-V: option requires Groove Name.")

            tfile = "MMAtmp%s.mma" % os.getpid()
            op = open(tfile, "w")
            groove = ''
            cmds = []
            chords = "I, vi, ii, V7"
            count = 4
            for g in args:
                if '=' in g:
                    c = g.split('=')
                    if c[0].upper() == 'CHORDS':
                        chords = c[1]
                    elif c[0].upper() == "COUNT":
                        count = c[1]
                        try:
                            count = int(count)
                        except:
                            error("-V: expecting integer for Count.")
                    else:
                        cmds.append(c)
                elif groove:
                    error("-V: Only one groove name permitted.")
                else:
                    groove = g
            if not groove:
                error("-V: no groove name specified.")

            op.write("Groove %s\n" % groove)
            for g in cmds:
                op.write("%s %s \n" % (g[0], g[1]))
            chords = chords.split(',')
            while len(chords) < count:
                chords += chords
            chords = chords[:count]
            for c in chords:
                op.write("%s\n" % c)

            op.close()

            mma.MMA.exits.files.append(tfile)

            args = [tfile]

        elif o=='-x':  # any one of some xtra, seldom used, options
            import mma.MMA.xtra
            mma.MMA.xtra.xoption(a, args)
            
        else:
            usage()      # unreachable??

    if internal:
        return

    # we have processed all the args. Should just have a filename left

    if len(args) > 1:
        usage("Only one input filename is permitted, %s given on command line." % len(args))

    if gbl.infile:
        usage("Input filename already assigned ... should not happen.")

    if args:
        gbl.infile = args[0]

    # if a single '-' is left on the cmd line user want stdin. We set the
    # the input filename to numeric 1 which can't be entered.

    if gbl.infile == '-':
        gbl.infile = 1
        
        if not gbl.outfile:
            import mma.MMA.debug   # circular dep. problem
            if not(mma.MMA.debug.noOutput):
                error("Input from STDIN specified. Use -f to set an output filename.")


def usage(msg=''):
    """ Usage message. """

    txt = [
        "MMA - Musical Midi Accompaniment",
        "  Copyright 2003-19, Bob van der Poel. Version %s" % gbl.version,
        "  Distributed under the terms of the GNU Public License.",
        "  Usage: mma [opts ...] INFILE [opts ...]",
        "",
        "Options:",
        " -b <n> Limit compilation to n1-n2 bars (comment numbers)",
        " -B <n> Like -b but for absolute bar numbers",
        " -c    display default Channel assignments",
        " -d    enable lots of Debugging messages",
        " -Dk   print list of MMA keywords",
        " -Dxl  eXtract Latex doc blocks from file",
        " -Dxh  eXtract HTML doc blocks from file",
        " -Dgh  extract HTML Groove doc",
        " -Dbo  extract text for browser app",
        " -Ds   extract sequence lists from file",
        " -e    show parsed/Expanded lines",
        " -f <file>  set output Filename",
        " -g    update Groove dependency database",
        " -G    create Groove dependency database",
        " -i <file> specify init (mmarc) file",
        " -I <plugin> print docs for plugin if available",
        " -II   skip premissions test for plugins (Dangerous!)",
        " -L    show order of bars processed",
        " -m <x> set Maxbars (default == 500)",
        " -M <x> set SMF to 0 or 1",
        " -n    No generation of midi output",
        " -o    show complete filenames when Opened",
        " -p    display Patterns as they are defined",
        " -P    play song (don't save) with player",
        " -r    display Running progress",
        " -s    display Sequence info during run",
        " -S <var[=data]>  Set macro 'var' to 'data'",
        " -T <tracks> Limit generation to specified tracks",
        " -v    display Version number",
        " -V <groove [options]> preview play groove",
        " -w    disable Warning messages",
        " -xCHORDS=<chord list> test listed chords for validity",
        " -xNOCREDIT disable MMA credits in Midi Meta track",
        " -xCHECKFILE=<filename> check chords in file",
        " -0    create sync at start of all channel tracks",
        " -1    create sync at end of all channel tracks",
        " -     a single hyphen signals to use STDIN instead of a file"]

    for a in txt:
        print(a)

    if msg:
        print("\n%s" % msg)
    
    sys.exit(1)


def setBarRange(v):
    """ Set a range of bars to compile. This is the -B/b option."""
    
    if gbl.barRange:
        error("Only one -b or -B permitted.")

    for ll in v.split(','):

        l = ll.split("-")

        if len(l) == 2:
            s, e = l
            try:
                s = int(s)
                e = int(e)
            except:
                usage("-B/b ranges must be integers, not '%s'." % l)

            for a in range(s, e + 1):
                gbl.barRange.append(str(a))

        elif len(l) == 1:
            try:
                s = int(l[0])
            except:
                usage("-B/b range must be an integer, not '%s'." % l[0])
            gbl.barRange.append(str(s))

        else:
            usage("-B/b option expecting N1-N2,N3... not '%s'." % v)



# lyric.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

from . import gbl
from   mma.MMA.common import *
import mma.MMA.paths
import mma.MMA.debug

class Lyric:
    textev = None    # set if TEXT EVENTS (not recommended)
    barsplit = None    # set if lyrics NOT split into sep. events for bar
    versenum = 1       # current verse number of lyric
    dupchords = 0       # set if we want chords as lyric events
    transpose = 0       # tranpose chord names (for dupchords only)
    karmode = 0       # in kar mode use textevents, split at hyphens
    enabled = True    # set true/false 
    pushedLyrics = []

    transNames = (('C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B'),
                 ('C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'))

    transKey = 0   # 0==flat, 1=sharp

    chordnames = {
        'B#': 0, 'C': 0, 'C#': 1, 'Db': 1,
        'D': 2, 'D#': 3, 'Eb': 3, 'E': 4,
        'Fb': 4, 'E#': 5, 'F': 5, 'F#': 6,
        'Gb': 6, 'G': 7, 'G#': 8, 'Ab': 8,
        'A': 9, 'A#': 10, 'Bb': 10, 'B': 11,
        'Cb': 11}

    def __init__(self):
        pass

    def setting(self):
        """ Called from macro. """

        a = "Event="

        if self.textev:
            a += "Text"
        else:
            a += "Lyric"

        a += " Split="
        if self.barsplit:
            a += "Bar"
        else:
            a += "Normal"

        a += " Verse=%s" % self.versenum

        a += " Chords="
        if self.dupchords:
            a += "On"
        else:
            a += "Off"

        a += " Transpose=%s" % self.transpose

        a += " CNames="
        if self.transKey:
            a += "Sharp"
        else:
            a += "Flat"

        a += " KAR="
        if self.karmode:
            a += "On"
        else:
            a += "Off"

        return a

    def option(self, ln):
        """ Set a lyric option. """

        ln, opts = opt2pair(ln)

        for o, v in opts:
            o = o.upper()
            v = v.upper()

            if o == 'EVENT':
                if v == 'TEXT':
                    self.textev = 1
                    warning("Lyric: Placing lyrics as TEXT EVENTS is not recommended")

                elif v == 'LYRIC':
                    self.textev = None
                    if mma.MMA.debug.debug:
                        dPrint("Lyric: lyrics set as LYRIC events.")

                else:
                    error("Lyric: Valid options for EVENT are TEXT or LYRIC.")

            elif o == 'SPLIT':
                if v == 'BAR':
                    self.barsplit = 1
                    if mma.MMA.debug.debug:
                        dPrint("Lyric: lyrics distributed thoughout bar.")

                elif v == 'NORMAL':
                    self.barsplit = None
                    if mma.MMA.debug.debug:
                        dPrint("Lyric: lyrics appear as one per bar.")

                else:
                    error("Lyric: Valid options for SPLIT are BAR or NORMAL.")

            elif o == 'VERSE':
                if v.isdigit():
                    self.versenum = int(v)

                elif v == 'INC':
                    self.versenum += 1

                elif v == 'DEC':
                    self.versenum -= 1

                else:
                    error("Lyric: Valid options for VERSE are <nn>, INC or DEC")

                if self.versenum < 1:
                    error("Lyric: Attempt to set Verse to %s. Values must be > 0" % self.versenum)

                if mma.MMA.debug.debug:
                    dPrint("Lyric: Verse number set to %s" % self.versenum)

            elif o == 'CHORDS':
                if v in ('1', 'ON'):
                    self.dupchords = 1
                    if mma.MMA.debug.debug:
                        dPrint("Lyric: Chords are duplicated as lyrics.")

                elif v in ('0', 'OFF'):
                    self.dupchords = 0
                    if mma.MMA.debug.debug:
                        dPrint("Lyric: Chords are NOT duplicated as lyrics.")

                else:
                    error("Lyric: CHORDS expecting 'ON' or 'OFF', not %s'" % v)

            elif o == 'TRANSPOSE':
                addTrans = False
                for t in v.split(','):
                    if t.upper() == 'ADD': # keyword
                        addTrans=True
                    else:  # either int or up-maj-3, etc.
                        v = mma.MMA.keysig.getTranspose([t], "Lyric Transpose")

                if addTrans:
                    self.transpose += v
                else:
                    self.transpose = v

                if mma.MMA.debug.debug:
                    dPrint("Lyric: Chord names transposed %s steps." % self.transpose)

            elif o == 'ADDTRANSPOSE':
                self.transpose += mma.MMA.keysig.getTranspose([v], "Lyric AddTranspose")

                if mma.MMA.debug.debug:
                    dPrint("Lyric: Chord names transposed %s steps." % self.transpose)

            elif o == 'CNAMES':
                if v in ('#', 'SHARP'):
                    self.transKey = 1

                elif v in ('B', '&', 'FLAT'):
                    self.transKey = 0

                else:
                    error("Lyric CNames: expecting 'Sharp' or 'Flat', not '%s'" % v )

                if mma.MMA.debug.debug:
                    msg = "Lyric: Chord names favor "
                    if self.transKey:
                        msg += "#."
                    else:
                        msg += "b."
                    dPrint(msg)

            elif o == 'KARMODE':
                if v in ('ON', '1'):
                    self.karmode = 1
                    if not hasattr(self, 'setkar'):
                        self.setkar = 1
                        meta = gbl.mtrks[0]
                        # this converts the "created" text to kar format
                        mt = meta.miditrk
                        if 0 in mt:   # don't bother if no events at 0
                            txt = None
                            for t, ev in enumerate(mt[0]):
                                if ev[1] == 1:
                                    ev = ev[3:]
                                    if ev.startswith(b"Created by MMA"):
                                        txt = b"@I " + ev
                                        del mt[0][t]
                                        break
                            if txt:
                                meta.addText(0, txt)
                        # other kar fields
                        meta.addText(0, b"@KMIDI KARAOKE FILE")
                        meta.addText(0, b"@V0100")
                        # change extension to .kar
                        mma.MMA.paths.createOutfileName('.kar')
 
                elif v in ('OFF', '0'):
                    self.karmode = 0
                else:
                    error("Lyric Kar: expecting On, 1, Off or 0, not '%s'." % v)

                if mma.MMA.debug.debug:
                    msg = "Lyric: Karmode",
                    if self.karmode:
                        msg += "enabled."
                    else:
                        msg += "disabled."
                    dPrint(msg)

            else:
                error("Usage: Lyric expecting EVENT, SPLIT, VERSE, CHORDS, TRANSPOSE,"
                      "CNAMES, KARMODE, ENABLE or SET, not '%s'" % o)

        # All the opt=value options have been taken care of. ln can now only
        # contain "On", "OFF" or "Set ..." Anything else is an error.

        while 1:
            if not ln:
                return

            if ln[0].upper() == 'OFF':
                enable = False
                ln.pop(0)

            elif ln[0].upper() == 'ON':
                enable = True
                ln.pop(0)

            elif ln[0].upper() == "SET":
                s = ' '.join(ln[1:]).strip()

                if not s.startswith('['):
                    s = '[' + s + ']'

                self.pushedLyrics.append(s)
                ln = []

            else:
                error("Lyric: Unknown option '%s'." % ln[0])

    def leftovers(self):
        """ Just report leftovers on stack."""

        if self.pushedLyrics:
            warning("Lyrics remaining on stack")

    def extract(self, ln, rpt):
        """ Extract lyric info from a chord line and place in META track.

            Returns line and lyric as 2 strings.

            The lyric is returned for debugging purposes, but it has been
            processed and inserted into the MIDI track.
        """

        a = ln.count('[')
        b = ln.count(']')

        if a != b:
            error("Mismatched []s for lyrics found in chord line")

        if self.pushedLyrics:
            if a or b:
                error("Lyrics not permitted inline and as LYRIC SET")

            ln = ln + self.pushedLyrics.pop(0)
            a = b = 1   # flag that we have lyrics, count really doesn't matter

        ln, lyrics = pextract(ln, '[', ']')

        # If the CHORDS=ON option is set, make a copy of the chords and
        # insert as lyric. This permits illegal chord lines, but they will
        # be caught by the parser. NOTE: Also discards [] or SET lyrics.

        if self.dupchords:
            ly = []
            for v in ln.split():   # do each chord symbol or '/' mark

                if v != '/':
                    v = v.replace('&', 'b')    # remember, "&" is optional

                    if v == 'z' or v == 'z!':  # convert plain z to "NC"
                        v = 'N.C.'

                    if 'z' in v:               # strip out the 'zCDA..' after the chord
                        v = v.split('z')[0]

                    v = v.lstrip("+-")         # strip off leading "+" and "-"

                    if self.transpose:   # transpose will be set to 0, 1, -1, etc.
                        t = None         # Needed in case line is not a chord ('/', "NC")!

                        try:             # try for 2 char chord name (F#, Gb, etc)
                            cn = v[0:2]
                            t = self.chordnames[cn] + self.transpose
                        except:
                            try:         # try 1 char chord name (C, D, etc)
                                cn = v[0:1]
                                t = self.chordnames[cn] + self.transpose
                            except:
                                pass     # not a chord pitch

                        if t is not None:    # use None, 0 is okay
                            while t >= 12:
                                t -= 12
                            while t <= -12:
                                t += 12

                            v = self.transNames[self.transKey][t] + v[len(cn):]

                ly.append(v)

            i = int(gbl.QperBar - len(ly))
            if i > 0:
                ly.extend(['/'] * i)
            lyrics.insert(0, ' '.join(ly) + '\\r')

        v = self.versenum

        if len(lyrics) == 1:
            v = 1

        if v > len(lyrics):
            lyrics = ''
        else:
            lyrics = lyrics[v-1]

        if not len(lyrics):
            return (ln, [])

        lyrics = lyrics.replace('\\r', ' \\r ')
        lyrics = lyrics.replace('\\n', ' \\n ')
        lyrics = lyrics.replace('     ', ' ')

        if self.karmode:
            lyrics = lyrics.replace('\-', chr(1))
            lyrics = lyrics.replace('-', chr(0)+' ')

        if self.barsplit:
            lyrics = [lyrics]
        else:
            lyrics = lyrics.split()

        beat = 0
        bstep = gbl.QperBar / float(len(lyrics))

        for t, a in enumerate(lyrics):  # do by syllable
            a, b = pextract(a, '<', '>', onlyone=True)

            if self.dupchords:
                a=a.split("@")[0]  # ignore chord position marker
            
            if b and b[0]:
                beat = stof(b[0], "Expecting value in <%s> in lyric" % b)
                if beat < 1 or beat > gbl.QperBar+1:
                    error("Offset in lyric <> must be 1 to %s" % gbl.QperBar)
                beat -= 1
                bstep = (gbl.QperBar-beat)/float((len(lyrics)-t))

            a = a.replace('\\r', '\r')
            a = a.replace('\\n', '\n')

            if a and a != ' ':
                if a and self.karmode and (chr(0) in a or chr(1) in a):
                    a = a.replace(chr(0), '')
                    a = a.replace(chr(1), '-')
                elif not a.endswith('-') and not a.endswith('\n') and not a.endswith('\r'):
                    a += ' '

                # Call midi module to add lyric syllable to the MIDI track
                p = getOffset(beat * gbl.BperQ)
                if self.enabled:
                    if self.textev or self.karmode:
                        gbl.mtrks[0].addText(p, a)
                    else:
                        gbl.mtrks[0].addLyric(p, a)

            beat += bstep
        return (ln, lyrics)

# Create a single instance of the Lyric Class.

lyric = Lyric()



# midiC.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module contains interface for MIDI constants and conversion routines.
"""

from mma.MMA.common import *
from mma.MMA.miditables import *
import mma.MMA.translate


def decodeVoice(n):
    """ Convert a single voice item. Used by setVoice() and tweaks. """

    n = mma.MMA.translate.vtable.get(n)
    voc = instToValue(n)

    if voc < 0 and n[0].isalpha():
        error("Voice '%s' is not defined." % n)
    if voc < 0:  # not a valid name, assume vv.msb(ctrl0).lsb(ctrl32) value
        nn = n.split('.')
        if len(nn) > 3 or len(nn) < 1:
            error("Expecting a voice value Prog.MSB.LSB, not '%s'" % n)
        voc = 0
        if len(nn) > 2:
            i = stoi(nn[2])
            if i < 0 or i > 127:
                error("LSB must be 0..127, not '%s'" % i)
            voc = i << 16

        if len(nn) > 1:
            i = stoi(nn[1])
            if i < 0 or i > 127:
                error("MSB must be 0..127, not '%s'" % i)
            voc += i << 8

        i = stoi(nn[0])
        if i < 0 or i > 127:
            error("Program must be 0..127, not '%s'" % i)
        voc += i

    return voc


def voice2tup(x):
    """ Convert integer into 3 byte tuple: Voice, LSB, MSB. """

    if x > 0xffff:
        msb = x >> 16
        x &= 0xffff
    else:
        msb = 0

    if x > 0xff:
        lsb = x >> 8
        x &= 0xff
    else:
        lsb = 0

    return (x, lsb, msb)


def extVocStr(v):

    v = "%s.%s.%s" % voice2tup(v)
    if v[-2:] == '.0':
        v = v[:-2]
    if v[-2:] == '.0':
        v = v[:-2]
    return v

###############################


def drumToValue(name):
    """ Get the value of the drum tone (-1==error).
        Note that this is quite different from instToValue() ... in that
        case the caller does a bunch of validation stuff for controllers, etc.
    """

    try:  # assuming that 'name' is an integer
        i = int(name, 0)
        if i >= 0 and i <= 127:
            return int(name)
        else:
            return -1
    except:
        try:
            return drumInx[name.upper()]
        except KeyError:
            return -1


def valueToDrum(val):
    """ Get the name of the drum tone.  """

    try:
        return drumNames[val]
    except KeyError:
        return str(val)


def instToValue(name):
    """ Get the value of the instrument name (-1==error). """

    try:
        return voiceInx[name.upper()]
    except KeyError:
        return -1


def valueToInst(val):
    """ Get the name of the inst. (or 'ERR'). """

    try:
        return voiceNames[val]
    except KeyError:
        try:
            int(val)
        except:
            return "ERROR"
        return extVocStr(val)


def ctrlToValue(name):
    """ Get the value of the controler name (-1==error). """

    try:
        return ctrlInx[name.upper()]
    except KeyError:
        return -1


def valueToCtrl(val):
    """ Get the name of the controller (or 'ERR'). """

    try:
        return ctrlNames[val]
    except KeyError:
        return "ERROR"



# midi.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>
"""


from   mma.MMA.midiM import intToWord, intTo3Byte, intToLong, intToVarNumber, intTo14, packBytes
import mma.MMA.midiC
import mma.MMA.debug
import mma.MMA.sync
from . import gbl
from   mma.MMA.common import *
from   mma.MMA.miditables import NONETONE

splitChannels = []

# some constants we use to catorgize event types
MIDI_NOTE = 1
MIDI_PRG = 2


def setSplitChannels(ln):
    """ Parser routine, sets up list of track to split. Overwrites existing. """

    global splitChannels
    splitChannels = []

    for a in ln:
        try:   # is this a legit channel number
            ch = int(a, 0)
        except ValueError:  # assume track name then
            ch = None

        if ch is None:
            if not a in gbl.tnames:
                a = a.upper()
                mma.MMA.alloc.trackAlloc(a, 0)    # ensure that track is allocated
            if not a in gbl.tnames:
                error("MidiSplit: Track '%s' is not valid." % a)
            if not gbl.tnames[a].channel:
                gbl.tnames[a].setChannel()
            ch = gbl.tnames[a].channel

        if ch < 1 or ch > 16:
            error("MidiSplit: Expecting value 1 to 16, not %s" % a)
        if ch not in splitChannels:
            splitChannels.append(ch)

    if mma.MMA.debug.debug:
        msg = ["SplitChannels: "]
        for a in splitChannels:
            msg.append(str(a))
        dPrint(' '.join(msg))


####################

def writeTracks(out):
    """ Write the accumulated MIDI tracks to file. """

    keys = list(gbl.mtrks.keys())
    keys.sort()

    """ For type 0 MIDI files all data is contained in 1 track.
        We take all our tracks and copy them to track 0, then
        set up keys[] so that only track 0 remains.
    """

    if gbl.midiFileType == 0:
        trk0 = gbl.mtrks[0].miditrk
        for n in keys[1:]:
            trk = gbl.mtrks[n].miditrk
            for k, v in trk.items():
                if k in trk0:
                    trk0[k].extend(v)
                else:
                    trk0[k] = v
        keys = [0]

    # Write header

    tcount = len(keys)
    out.write(mkHeader(tcount, gbl.BperQ, gbl.midiFileType))

    if gbl.barRange:   # compensate for -B/-b options
        stripRange()

    # Write data chunks for each track

    for n in keys:
        if n in splitChannels and gbl.midiFileType:
            tcount += writeSplitTrack(n, out)
        else:
            gbl.mtrks[n].writeMidiTrack(out)

    """ We may have changed the track count! So, we need to
        fix the file header. This is offset 10/11 which contains
        the number of tracks. The counter tcount has been
        tracking this, so just seek, replace and seek back.
    """

    if tcount != len(keys):
        out.seek(0)
        out.write(mkHeader(tcount, gbl.BperQ, gbl.midiFileType))
        out.seek(0, 2)  # return to eof


def writeSplitTrack(channel, out):
    """ Split a track. In drum tracks this puts different instruments
        into individual tracks (useful!); for instrument tracks it puts
        each pitch into a track (probably not useful).
    """

    tr = gbl.mtrks[channel].miditrk   # track to split

    # A dict to store the split midi tracks. We'll end out with
    # a track for each pitch which appears in the track and
    # a track (labeled -1) to store every other-than-note-on/off data.

    notes = {}

    onEvent = 0x90 + (channel-1)
    offEvent = 0x80 + (channel-1)
    for offset in tr.keys():
        for x in range(len(tr[offset])-1, -1, -1):
            ev = tr[offset][x]
            if len(ev) == 3 and (ev[0] in (onEvent, offEvent)):
                n = ev[1]
            else:
                n = -1      # special value for non-note on events

            if not n in notes:   # create a new mtrk if needed
                notes[n] = Mtrk(10)

            if offset in notes[n].miditrk:  # copy event to new track
                notes[n].miditrk[offset].append(ev)
            else:
                notes[n].miditrk[offset] = [ev]

    if mma.MMA.debug.debug:
        dPrint(" Data has been split into %s tracks." % len(notes))

    # Insert a channel name in all the new tracks.

    for a in notes.keys():
        if a == -1:
            continue
        if channel == 10:
            m = "%s" % mma.MMA.midiC.valueToDrum(a)
        else:
            m = "%s-%s" % (gbl.mtrks[channel].trackname, a)

        notes[a].addTrkName(0, m)

    for a in sorted(notes.keys()):
        notes[a].writeMidiTrack(out)

    """ The split tracks have been written. Return the number of additional tracks
        so that the caller can properly update the midi file header. Note that
        len(notes)-1 IS CORRECT ... we've already figured on writing 1 track.
    """

    return len(notes)-1


def mkHeader(count, tempo, Mtype):

    return packBytes("MThd", intToLong(6), intToWord(Mtype), intToWord(count), intToWord(tempo))


# Midi track class. All the midi creation is done here.
#  We create a class instance for each track. mtrks{}.

class Mtrk:

    def __init__(self, channel):
        self.miditrk = {}
        self.channel = channel-1
        self.trackname = ''
        self.lastOffEvent = [None] * 129  # cell for each note, saved delta
        self.lastPrg = 0

    def delDup(self, offset, cmd):
        """ Delete a duplicate event. Used by timesig, etc.    """

        tr = self.miditrk
        lg = len(cmd)
        if offset in tr:
            tmp = []
            for ev in tr[offset]:
                if ev[0:lg] != cmd:  # not target, copy it
                    tmp.append(ev)
            tr[offset] = tmp           # new set of events (may be empty)

    def addTimeSig(self, offset,  nn, dd, cc, bb):
        """ Create a midi time signature.

            delta - midi delta offset
            nn = sig numerator, beats per measure
            dd - sig denominator, 2=quarter note, 3=eighth,
            cc - midi clocks/tick
            bb - # of 32nd notes in quarter (normally 8)

            This is only called by timeSig.create(). Don't
            call this directly since the timeSig.create() checks for
            duplicate settings.
        """

        cmd = packBytes(0xff, 0x58)
        # we might have several different timesigs on the same offset,
        # so take time to delete any. 
        self.delDup(offset, cmd)
        self.addToTrack(offset, packBytes(cmd, (0x04, nn, dd, cc, bb)))

    def addKeySig(self, offset, n, mi):
        """ Set the midi key signature. """

        cmd = packBytes(0xff, 0x59)
        self.delDup(offset, cmd)
        self.addToTrack(offset, packBytes(cmd, (0x02, n, mi)))

    def addMarker(self, offset, msg):
        """ Create a midi MARKER event."""

        self.addToTrack(offset, packBytes((0xff, 0x06), intToVarNumber(len(msg)), msg))

    def addCopyright(self, offset, msg):
        """ Insert copyright. """

        # should never happen since the caller sets offset=0
        if offset != 0:
            error("Copyright message must be at offset 0, not %s." % offset)

        # We need to bypass addToTrack to force copyright to the start of the track.

        # Create the copyright event
        ev = packBytes((0xff, 0x02), intToVarNumber(len(msg)), msg)

        tr = self.miditrk   # this is the meta track

        # We keep a pointer (ipoint) which points to the position of
        # the last copyright string. If there isn't one, we create
        # it in the expect; else it's just incremented

        try:
            self.ipoint += 1
        except AttributeError:
            self.ipoint = 0

        if offset in tr:
            tr[offset].insert(self.ipoint, ev)
        else:
            tr[offset] = [ev]

    def addText(self, offset, msg):
        """ Create a midi TextEvent."""

        self.addToTrack(offset, packBytes((0xff, 0x01), intToVarNumber(len(msg)), msg))

    def addLyric(self, offset, msg):
        """ Create a midi lyric event. """

        self.addToTrack(offset, packBytes((0xff, 0x05), intToVarNumber(len(msg)), msg))

    def addCuePoint(self, offset, msg):
        """ Create a MIDI cue pointr event. """

        self.addToTrack(offset, packBytes((0xff, 0x07), intToVarNumber(len(msg)), msg ))

    def addTrkName(self, offset, msg):
        """ Creates a midi track name event. """

        offset = 0  # ignore user offset, always put this at 0

        self.trackname = msg

        cmd = packBytes((0xff, 0x03))
        self.delDup(offset, cmd)
        self.addToTrack(offset, packBytes(cmd, intToVarNumber(len(msg)), msg))

    def addProgChange(self, offset, program, oldprg):
        """ Create a midi program change (handles extended voicing).

            program - The MIDI program (voice) value
            oldprg  - existing MIDI program
        """

        # We truck around a special pseudo voice 'NONE' or 127.127.127 which
        # is a signal that we don't want mma to set the voicing. Might be
        # useful when know the track that mma is using and we have preset
        # an external synth.

        if program == NONETONE:
            return

        v1, lsb1, msb1 = mma.MMA.midiC.voice2tup(oldprg)
        v2, lsb2, msb2 = mma.MMA.midiC.voice2tup(program)

        if msb1 != msb2:   # only if CTRL32 has changed
            self.addToTrack(offset, packBytes((0xb0 | self.channel, 0x20, msb2)))

        if lsb1 != lsb2:   # only if CTRL0 has changed
            self.addToTrack(offset, packBytes((0xb0 | self.channel, 0x00, lsb2)))

        # Always do voice change. Maybe not necessary, but let's be safe.

        self.addToTrack(offset, packBytes((0xc0 | self.channel, v2)), MIDI_PRG)

    def addGlis(self, offset, v):
        """ Set the portamento. LowLevel MIDI.

            This does 2 things:
                1. turns portamento on/off,
                2. sets the LSN rate.
        """

        if v == 0:
            self.addToTrack(offset, packBytes((0xb0 | self.channel, 0x41, 0x00)))

        else:
            self.addToTrack(offset, packBytes((0xb0 | self.channel, 0x41, 0x7f)))
            self.addToTrack(offset, packBytes((0xb0 | self.channel, 0x05, v)))

    def addWheel(self, offset, v):
        """ Set lsb/msb for the modulation wheel. """

        self.addToTrack(offset, packBytes((0xe0 | self.channel), intTo14(v)))

    def addPan(self, offset, v):
        """ Set the lsb of the pan setting."""

        self.addToTrack(offset, packBytes((0xb0 | self.channel, 0x0a, v)))

    def addCtl(self, offset, l):
        """ Add arbitary control sequence to track."""

        self.addToTrack(offset, packBytes(0xb0 | self.channel, l))

    def addNoteOff(self, offset):
        """ Insert a "All Note Off" into the midi stream.

            Called from the cutTrack() function.
        """

        self.addToTrack(offset, packBytes((0xb0 | self.channel, 0x7b, 0)), MIDI_NOTE )

    def addMasterVolume(self, offset, v):
        """ System Exclusive master volume message. Meta track. """

        # We send to packBytes as a long list, no tuples. Just a bit
        # easier to maintain a long list like this.
        self.addToTrack(offset, packBytes(
                0xf0,       # Start sysex
                0x07,       # message size (needed for SMF)
                0x7f,       # realtime
                0x7f,       # disregard channel
                0x04,       # device control
                0x01,       # master volume
                intTo14(v), # params (14 bit)
                0xf7 ))     # EOX

    def addChannelVol(self, offset, v):
        """ Set the midi channel volume."""

        tr = self.miditrk
        cvol = packBytes(0xb0 | self.channel, 0x07)  # 2 byte channel vol

        # Before adding a new channel volume event we check to see if there
        # are any future channel volume events and delete them.
        
        for off in tr:
            if off >= offset:
                tr[off] = [e for e in tr[off] if e[0:2] != cvol]
        
        self.addToTrack(offset, packBytes(cvol, v))

    def addTempo(self, offset, beats):
        """ Create a midi tempo meta event.

             beats - beats per second

             Return - packed midi string
        """

        cmd = packBytes((0xff, 0x51, 0x03))
        self.delDup(offset, cmd)

        self.addToTrack(offset, packBytes(cmd, intTo3Byte(60000000 // beats)))

    def writeMidiTrack(self, out):
        """ Create/write the MIDI track.

            We convert timing offsets to midi-deltas.
        """

        tr = self.miditrk

        """ If the -1 command line option is set we need to add a terminate
            to the end of each track. This is done to make looping
            software like seq24 happy. We do this by truncating all
            data in the file past the current tick pointer and inserting
            an all-notes-off at that position.
        """

        if mma.MMA.sync.endsync and self.channel >= 0:
            eof = gbl.tickOffset
            for offset in tr.keys():
                if offset > eof:
                    del tr[offset]
            self.addToTrack(eof, packBytes((0xb0 | self.channel, 0x7b, 0)))

        """ To every MIDI track we generate we add (if the -0 flag
            was set) an on/off beep at offset 0. This makes for
            easier sync in multi-tracks.
        """

        if mma.MMA.sync.synctick and self.channel >= 0:
            t, v = mma.MMA.sync.syncTone
            self.addToTrack(0, packBytes((0x90 | self.channel, t, v)))
            self.addToTrack(1, packBytes((0x90 | self.channel, t, 0)))

        if mma.MMA.debug.debug:
            ttl = 0
            lg = 1
            for t in tr:
                a = len(tr[t])
                if a > lg:
                    lg = a
                ttl += a
            if self.channel == -1:
                nm = "META"
            else:
                nm = self.trackname
            dPrint( "<%s> Unique ts: %s; Ttl events %s; Average ev/ts %.2f" %
                (nm, len(tr), ttl, float(ttl)/len(tr)))

        last = 0

        # Convert all events to MIDI deltas and store in
        # the track array/list

        tdata = []        # empty track container
        lastSts = None    # Running status tracker

        for a in sorted(tr.keys()):
            delta = a-last

            if not tr[a]:
                continue  # this skips the delta offset update!

            for d in tr[a]:
                """ Running status check. For each packet compare
                    the first byte with the first byte of the previous
                    packet. If it is can be converted to running status
                    we strip out byte 0. Note that valid running status
                    byte are 0x80..0xef. 0xfx are system messages
                    and are not suitable for running status.
                """

                if len(d) > 1:
                    if d[0] == lastSts:
                        d = d[1:]
                    else:
                        lastSts = d[0]
                        if lastSts < 0x80 or lastSts > 0xef or not gbl.runningStatus:
                            lastSts = None

                tdata.extend([intToVarNumber(delta), d])
                delta = 0
            last = a

        # Add an EOF to the track (included in total track size)

        tdata.append(intToVarNumber(0))
        tdata.append(packBytes((0xff, 0x2f, 0x00)))

        tdata = bytearray(b'').join(tdata)
        totsize = len(tdata)

        out.write(b"MTrk")
        out.write(intToLong(totsize))
        out.write(tdata)

    def addPairToTrack(self, boffset, startRnd, endRnd, duration, note, v, unify ):
        """ Add a note on/off pair to a track.

            boffset      - offset into current bar
            startRnd, endRnd  - rand val start adjustment
            duration  - note len
            note      - midi value of note
            v      - midi velocity
            unify      - if set attempt to unify/compress on/offs

            This function tries its best to handle overlapping events.
            Easy to show effect with a table of note ON/OFF pairs. Both
            events are for the same note pitch.

            Offsets     |     200  |      300  |  320  |  420
            ---------|--------|--------|-------|--------
            Pair1     |     on      |       |  off  |
            Pair2     |      |      on   |       |  off

            The logic here will delete the OFF event at 320 and
            insert a new OFF at 300. Result is that when playing
            Pair1 will turn off at 300 followed by the same note
            in Pair2 beginning sounded right after. Why the on/off?
            Remember: Velocities may be different!

            However, if the unify flag is set we should end up with:

            Offsets     |     200  |      300  |  320  |  420
            ---------|--------|--------|-------|--------
            Pair1     |     on      |       |       |
            Pair2     |      |       |       |  off

        """

        # Start/end offsets

        onOffset = getOffset(boffset, startRnd, endRnd)

        # If the ON or OFF offset is <0 we change it to 0 for a couple of
        # reasons:
        #   1. It'll be converted anyway in 'addToTrack', but that doesn't
        #      get back to us here.
        #   2. If we don't that wrong offset will be reported in lastOffEvent.

        if onOffset < 0:
            onOffset = 0
        offOffset = onOffset + duration
        if offOffset < 0:
            offOffset = 0

        # ON/OFF events

        onEvent = packBytes(0x90 | self.channel, note, v)
        offEvent = packBytes(onEvent[:-1], 0)

        noOnFlag = False

        f = self.lastOffEvent[note] 

        if f is not None and f >= onOffset and f <= offOffset:
            # evlist is a delta-offset list. It should have a note off event
            # for the this note. Just in case, we do check; but it's probably
            # not necessary. The off event is deleted and the unify stuff is done
            if offEvent in self.miditrk[f]:
                self.miditrk[f].remove(offEvent)
                if not unify:
                    self.addToTrack(onOffset, offEvent, MIDI_NOTE)
                else:
                    noOnFlag = True

        if not noOnFlag:
            self.addToTrack(onOffset, onEvent, MIDI_NOTE)
        self.addToTrack(offOffset, offEvent, MIDI_NOTE)

        # Save the NOTE OFF time for the next loop.

        self.lastOffEvent[note] = offOffset

    def addNoteOnToTrack(self, boffset, note, v, startRnd=None, endRnd=None):
        """ Add a single note on or note off when v=0 to a track.
            boffset      - offset into current bar
            duration  - note len
            note      - midi value of note
            v      - midi velocity, set to 0 for note off
            startRnd/endRnd  - rand val start adjustment

            Added by louisjb for plectrum.
        """

        # Start offsets

        onOffset = getOffset(boffset, startRnd, endRnd)

        # ON/OFF events (off is on with v = 0)

        onEvent = packBytes((0x90 | self.channel, note, v))

        self.addToTrack(onOffset, onEvent, MIDI_NOTE)
        return onOffset

    def addToTrack(self, offset, event, evType=None):
        """ Add an event to a track.

            MIDI data is saved as created in track structures.
            Each track has a miditrk dictionary entry which used
            the time offsets and keys and has the various events
            as data. Each event is a packed string of bytes and
            the events are stored as a list in the order they are
            created. Our storage looks like:

                 miditrk[OFFSET_VALUE] = [event1, event2, ...]

            evType is an optional arg. Two values are used:
                 MIDI_PR - a program (voice) change. Save the timestamp.
                 MIDI_NOTE - note on/off ... check to see it doesn't happen
                             before the last program change.

        """

        if evType == MIDI_NOTE and offset < self.lastPrg:
            offset = self.lastPrg
        elif evType == MIDI_PRG:
            self.lastPrg = offset

        # Important. It's possible that a negative delay could generate
        # a negative offset.

        if offset < 0:
            offset = 0

        tr = self.miditrk

        if offset in tr:
            tr[offset].append(event)
        else:
            tr[offset] = [event]



def stripRange():
    """ Strip out range limited data. Only of -B/b option. """

    bp = gbl.barPtrs   # list generated at compile time

    if gbl.barRange[-1] == 'ABS':
        gbl.barRange.pop()  # delete abs marker (set by -B)
        for a in bp:        # convert comment numbers to abs numbers
            bp[a][0] = str(a)
    validRange = []

    for a in gbl.barRange:   # list of bars we want to produce
        for b in bp:
            if a == bp[b][0]:
                validRange.append([bp[b][1], bp[b][2]])

    if not validRange:
        dPrint("   Range directive -b/B would result in empty file.\n"
              "   Entire file is being created. Check the range.")
        return

    # Collaspe/merge the valid range pointers
    validRange.sort()   # barptrs was dict, so this list is not in order
    tmp = []
    a, b = validRange[0]
    for i in range(1, len(validRange)):
        if b+1 == validRange[i][0]:
            b = validRange[i][1]
        else:
            tmp.append([a, b])
            a, b = validRange[i]
    tmp.append([a, b])
    validRange = tmp  # list of event times to keep

    """ Create list of event times to discard. Each item in the list:
        is [ start-time, end-time]
     """

    disList = []
    lowestEv = bp[1][1]
    for a in validRange:
        disList.append([lowestEv, a[0]-1])
        lowestEv = a[1]

    """ Determine the last event time in the buffer. This is not nesc. the
        same as the last barptr ... tempos and midinote can write past the
        last bar. We need to test each track and find the highest event time.
    """

    lastev = -1
    for a in gbl.mtrks:
        z = sorted(gbl.mtrks[a].miditrk)
        if z[-1] > lastev:
            lastev = z[-1]

    disList.append([validRange[-1][1]+1, lastev])

    """ 1st pass. For each track take all the cut parts and adjust their
        offsets to the end offset of the previous keep section. Strip out
        all note on events, lyric and text events in the effected bits.
    """

    for start, end in disList:
        for n in gbl.mtrks:
            tr = gbl.mtrks[n].miditrk
            newEvents = []
            for ev in sorted(tr.keys()):   # sort is important!
                if ev >= start and ev <= end:
                    for e in tr[ev]:
                        e0 = e[0] & 0xf0
                        # skip note on events
                        if e0 == 0x90 and e[2]:
                            continue
                        # skip lyric and text events
                        if e[0] == 0xff:
                            if e[1] == 0x05 or e[1] == 0x01:
                                continue
                        # skip if event is note off and it already is present
                        # Note: mma never generates a 0x80 (note-off) event, but
                        # the test is included here since midi-inc may have them.
                        if (e0 == 0x90 and e[2] == 0) or (e0 == 0x80):
                            for x in newEvents:
                                if x == e:
                                    e = None
                                    break
                        if e:
                            newEvents.append(e)
                    del (tr[ev])

            """ We now have a new list of events for the 'start' offset (it might
                just be an empty list) AND we have deleted the events for 'start'
                to 'end'. Just a matter of inserted a new event list.
            """

            if newEvents:
                tr[start] = newEvents

    """ 2nd pass. Adjust offsets of stuff to keep. """

    offset = 0
    for vals in range(len(validRange)):   # each valid range
        start = validRange[vals][0]
        end = validRange[vals][1]+1
        offset += (start - disList[vals][0])
        for a in gbl.mtrks:               # each track
            tr = gbl.mtrks[a].miditrk
            for ev in sorted(tr.keys()):  # each event list
                if ev >= start and ev <= end:
                    newoffset = ev-offset
                    if ev != newoffset:  # don't append/copy creating duplicates
                        if newoffset in tr:
                            tr[newoffset].extend(tr[ev])
                        else:
                            tr[newoffset] = tr[ev]
                        del tr[ev]

    """ After all this compressing the end of file marker is past the
        end of the real data. This could bugger up the -1 option, so fix
        it. Go though all the tracks again and find the end.
    """

    lastev = -1
    for a in gbl.mtrks:
        z = sorted(gbl.mtrks[a].miditrk)
        if z and z[-1] > lastev:
            lastev = z[-1]
    gbl.tickOffset = lastev

    dPrint("  File has been truncated for -b/-B option. Bar count/time incorrect.")



#!/usr/bin/env python
import sys
import os
import os.path
from setuptools import setup
import appdirs

tests_require = []
if sys.version_info < (2, 7):
    tests_require.append("unittest2")


def read(fname):
    inf = open(os.path.join(os.path.dirname(__file__), fname))
    out = "\n" + inf.read().replace("\r\n", "\n")
    inf.close()
    return out


setup(
    name='appdirs',
    version=appdirs.__version__,
    description='A small Python module for determining appropriate " + \
        "platform-specific dirs, e.g. a "user data dir".',
    long_description=read('README.rst') + '\n' + read('CHANGES.rst'),
    classifiers=[c.strip() for c in """
        Development Status :: 4 - Beta
        Intended Audience :: Developers
        License :: OSI Approved :: MIT License
        Operating System :: OS Independent
        Programming Language :: Python :: 2
        Programming Language :: Python :: 2.6
        Programming Language :: Python :: 2.7
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.2
        Programming Language :: Python :: 3.3
        Programming Language :: Python :: 3.4
        Programming Language :: Python :: Implementation :: PyPy
        Programming Language :: Python :: Implementation :: CPython
        Topic :: Software Development :: Libraries :: Python Modules
        """.split('\n') if c.strip()],
    test_suite='test.test_api',
    tests_require=tests_require,
    keywords='application directory log cache user',
    author='Trent Mick',
    author_email='trentm@gmail.com',
    maintainer='Trent Mick; Sridhar Ratnakumar',
    maintainer_email='trentm@gmail.com; github@srid.name',
    url='http://github.com/ActiveState/appdirs',
    license='MIT',
    py_modules=["appdirs"],
)



# seqrnd.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import random

from mma.MMA.common import *
import mma.MMA.debug

""" SeqRnd variable is a list. The first entry is a flag:(0, 1 or x):
      0 - not set
      1 - set
      2 - set for specific tracks, track list starts at position [1]
"""

seqRnd = [0]       # set if SEQRND has been set
seqRndWeight = [1]


def setseq():
    """ Set up the seqrnd values, called from parse loop.


        returns:
           0...  a random sequence number,
           -1    signals that rndseq is not enabled.

        There are three methods of rndseq. They depend on the first value
        in the list seqRnd[]:

           [0]  not enabled.
           [1]  random selection, keeps all tracks in sync.
           [2]  randomize selected tracks.

        The list seqRndWeight has at least the same number of entries in it
        as there are in the sequence size. By default each will have
        the same value of 1. A list of the valid sequence points is
        generated (ie, if seqsize==4 tmp will be [0,1,2,3]). In addition
        the weights of each entry in tmp[] is adjusted by the weights in
        the seqrndweight[] list.
    """

    if seqRnd[0]:
        r = getrndseq(seqRndWeight)
        if seqRnd[0] == 1:
            gbl.seqCount = r
            r = -1
    else:
        r = -1

    return (r, seqRnd[1:])


def getrndseq(v):
    tmp = []
    for x, i in enumerate(v):
        tmp.extend([x] * i)

    if not len(tmp):
        error("SeqRndWeight has generated an empty list")

    return random.choice(tmp)


## Main parser routines

def setSeqRnd(ln):
    """ Set random order for all tracks. """

    global seqRnd
    
    emsg = "use [ON, OFF or TrackList ]"
    if not ln:
        error("SeqRnd:" + emsg)

    a = ln[0].upper()

    if a in ("ON", "1") and len(ln) == 1:
        seqRnd = [1]

    elif a in ("OFF", "0") and len(ln) == 1:
        seqRnd = [0]

    else:
        seqRnd = [2]
        for a in ln:
            a = a.upper()
            if not a in gbl.tnames:
                error("SeqRnd: Track '%s' does not exist, %s" % (a, emsg))
            if a in seqRnd:
                error("SeqRnd: Duplicate track '%s' specified, %s" % (a, emsg))
            seqRnd.append(a)

    if mma.MMA.debug.debug:
        msg = ["SeqRnd:"]
        if seqRnd[0] == 2:
            for a in seqRnd[1:]:
                msg.append(a)
        elif seqRnd[0] == 1:
            msg.append("On")
        else:
            msg.append("Off")
        dPrint(' '.join(msg))


def getweights(ln, msg):

    ln = lnExpand(ln, msg)

    if not ln:
        error("Use: %s <weight factors>" % msg)

    tmp = []
    for n in ln:
        n = stoi(n)
        if n < 0:
            error("%s: Values must be 0 or greater" % msg)
        tmp.append(n)

    tmp = seqBump(tmp)

    if mma.MMA.debug.debug:
        dPrint("%s: %s" % (msg, ' '.join([str(x) for x in tmp])))

    return tmp


def setSeqRndWeight(ln):
    """ Set global rnd weight. """

    global seqRndWeight

    seqRndWeight = getweights(ln, "SeqRndWeight")



# harmony.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

from mma.MMA.common import *
import mma.MMA.debug


def setHarmony(self, ln):
    """ Set the harmony. """

    ln = lnExpand(ln, '%s Harmony' % self.name)
    tmp = []

    for n in ln:
        n = n.upper()
        if n in ('-', '0', 'NONE'):
            n = None

        tmp.append(n)

    self.harmony = seqBump(tmp)

    if self.vtype in ('CHORD', 'DRUM'):
        warning("Harmony setting for %s track ignored" % self.vtype)

    if mma.MMA.debug.debug:
        mma.MMA.debug.trackSet(self.name, "Harmony")

        
def setHarmonyOnly(self, ln):
    """ Set the harmony only. """

    ln = lnExpand(ln, '%s HarmonyOnly' % self.name)
    tmp = []

    for n in ln:
        n = n.upper()
        if n in ('-', '0', 'NONE'):
            n = None

        tmp.append(n)

    self.harmony = seqBump(tmp)
    self.harmonyOnly = seqBump(tmp)

    if self.vtype in ('CHORD', 'DRUM'):
        warning("HarmonyOnly setting for %s track ignored" % self.vtype)

    if mma.MMA.debug.debug:
        mma.MMA.debug.trackSet(self.name, 'HarmonyOnly')


def setHarmonyVolume(self, ln):
    """ Set harmony volume adjustment. """

    ln = lnExpand(ln, '%s HarmonyOnly' % self.name)
    tmp = []

    for n in ln:
        v = stoi(n)

        if v < 0:
            error("HarmonyVolume adjustment must be positive integer")
        tmp.append(v/100.)

    self.harmonyVolume = seqBump(tmp)

    if self.vtype in ('PLECTRUM', 'DRUM'):
        warning("HarmonyVolume adjustment for %s track ignored" % self.vtype)

    if mma.MMA.debug.debug:
        mma.MMA.debug.trackSet(self.name, "HarmonyVolume")


##########################################################

def harmonize(hmode, note, chord):
    """ Get harmony note(s) for given chord. """

    hnotes = []

    for tp in hmode.split('+'):

        if tp in ('2', '2BELOW'):
            hnotes.append(gethnote(note, chord))

        elif tp == '28Below':
            hnotes.append(gethnote(note, chord)-12)

        elif tp == '2ABOVE':
            hnotes.append(gethnote(note, chord)+12)

        elif tp == '28ABOVE':
            hnotes.append(gethnote(note, chord)+24)

        elif tp in ('3', '3BELOW'):
            a = gethnote(note, chord)
            b = gethnote(a, chord)
            hnotes.extend([a, b])
            
        elif tp == '38BELOW':
            a = gethnote(note, chord)
            b = gethnote(a, chord)
            hnotes.extend([a-12, b-12])

        elif tp == '3ABOVE':
            a = gethnote(note, chord)
            b = gethnote(a, chord)
            hnotes.extend([a+12, b+12])

        elif tp == '38ABOVE':
            a = gethnote(note, chord)
            b = gethnote(a, chord)
            hnotes.extend([a+24, b+24])
            
        elif tp in ('OPEN', "OPENBELOW"):
            a = gethnote(note, chord)
            hnotes.append(gethnote(a, chord))

        elif tp == 'OPEN8BELOW':
            a = gethnote(note, chord)
            hnotes.append(gethnote(a-12, chord))

        elif tp == 'OPENABOVE':
            a = gethnote(note, chord)
            hnotes.append(gethnote(a, chord) + 12)
 
        elif tp == 'OPEN8ABOVE':
            a = gethnote(note, chord)
            hnotes.append(gethnote(a, chord) + 24)

        elif tp in ('8', '8BELOW'):
            hnotes.append(note - 12)

        elif tp == '8ABOVE':
            hnotes.append(note + 12)

        elif tp in ('16', '16BELOW'):
            hnotes.append(note - (2 * 12))

        elif tp == '16ABOVE':
            hnotes.append(note + (2 * 12))

        elif tp in ('24', '24BELOW'):
            hnotes.append(note - (3 * 12))

        elif tp == '24ABOVE':
            hnotes.append(note + (3 * 12))

        elif ":" in tp:
            hnotes.append(note + intervalHarmony(tp))
            
        else:
            error("Unknown harmony type '%s'" % tp)

    # Strip out duplicate notes from harmony list.
    #  Cute trick here ... just use set().

    return list(set(hnotes))


# The following are used for the interval harmonies (2:Per4, etc)
# Whole tone to half step conversion table

halfSteps = { 'UNI':  0,
              'MIN2': 1,
              'MAJ2': 2,  'DIM3': 2,
              'MIN3': 3,  'AUG2': 3,
              'MAJ3': 4,  'DIM4': 4,
              'PER4': 5,  'AUG3': 5,
              'AUG4': 6,  'DIM5': 6,
              'PER5': 7,  'DIM6': 7,
              'MIN6': 8,  'AUG5': 8,
              'MAJ6': 9,  'DIM7': 9,
              'MIN7': 10, 'AUG6': 10, 
              'MAJ7': 11, 'DIM8': 11,
              'OCT':  12, 'AUG7': 12  }

sNames = ("MINOR", "MAJOR", "DIMINISHED", "PERFECT", "AUGMENTED", "OCTAVE", "UNISON")
dNames = (("SECOND", "2"), ("THIRD", "3"), ("FOURTH", "4"),
           ("FIFTH", "5"), ("SIXTH", "6"), ("SEVENTH", "7") )

def intervalHarmony(harmName):
    """ Return number of 1/2 steps for harmony. Syntax is:
            [octave]:interval
              octave - integer -5 to +5 (optional)
              interval - symbolic name or value
    """

    # since the caller got here due to a ':' in the string, we're
    # safe in doing the split.
    octave, name = harmName.split(':', 1)

    # set the octave
    if octave:
        octave = stoi(octave, "Harmony: octave expecting integer")
        if octave <-4 or octave>4:
            error("Harmony: Octave %s is too large. Use -4 to 4" % octave)
    else:
        octave = 0

    if name.isdigit():
        halfs = stoi(name, "Harmony: Unexpected error in number of half steps.")

    else:
        # Convert full names to our abreviations. This permits a
        # wide range of input names and fairly short lookup table.
        for a in sNames:
            name = name.replace(a, a[:3])
        for a in dNames:
            name = name.replace(a[0], a[1])

        if name not in halfSteps:
            error("Interval harmony %s is unknown." % harmName)

        halfs = halfSteps[name]
        
    return (octave * 12) + halfs


def gethnote(note, chord):
    """ Determine harmony notes for a note based on the chord.

        note - midi value of the note

        chord - list of midi values for the chord


        This routine works by creating a chord list with all
        its notes having a value less than the note (remember, this
        is all in MIDI values). We then grab notes from the end of
        the chord until one is found which is less than the original
        note.
    """

    wm = "No harmony note found since no chord, using note " \
         "0 which will sound bad"

    if not chord:        # should never happen!
        warning(wm)
        return 0

    ch = list(chord)    # copy chord and sort
    ch.sort()

    # ensure that the note is in the chord

    while ch[-1] < note:
        for i, n in enumerate(ch):
            ch[i] += 12

    while ch[0] >= note:
        for i, v in enumerate(ch):
            ch[i] -= 12

    # get one lower than the note

    while 1:
        if not ch:            # this probably can't happen
            warning(wm)
            return 0

        h = ch.pop()
        if h < note:
            break

    return h




# parseCL.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


This module parses a chordline and sets the timing offsets for 
each "tab". 

"""

from . import gbl
from mma.MMA.common import *

import mma.MMA.truncate

# This table is passed to the track classes. It has
# an instance for each chord in the current bar.

class CTable:
    name = None        # Chord name (used by plectrum track)
    chord = None       # A pointer to the chordNotes structures
    lastchord = None   # chord from previous ctable. Used by trigger()
    chStart = None     # where in the bar the chord starts (in ticks, 0..)
    chEnd = None       # where it ends (in ticks)
    chordZ = None      # set if chord is tacet
    arpeggioZ = None   # set if arpeggio is tacet
    walkZ = None       # set if walking bass is tacet
    drumZ = None       # set if drums are tacet
    bassZ = None       # set if bass is tacet
    scaleZ = None      # set if scale track is tacet
    ariaZ = None       # set if aria track is tacet
    plectrumZ = None   # set if plectrum is tacet

lastChord = None   # tracks last chord for "/ /" data lines.

chordTabs = []  # initialized by mma.MMA.main call to setTime()

def setChordTabs(l):
    """ Set the tab positions for chord parsing.

        This table is set up on each TIME change. Need to
        do it even if the TIME is NOT changed since the table
        positions might. Cheap to do, so not worth worrying.
    """

    global chordTabs

    chordTabs = tuple(( int((x-1) * gbl.BperQ) for x in l ))

def parseChordLine(l):
    """ Parse a line of chord symbols and determine start/end points. """

    global lastChord

    ctable = []               # an entry for each chord in the bar
    quarter = gbl.BperQ       # ticks in a quarter note (== 1 beat)
    if mma.MMA.truncate.length:
        endTime = mma.MMA.truncate.length
    else:
        endTime = (quarter * gbl.QperBar)  # number of ticks in bar

    p = 0                    # our beat counter --- points to tab 0, 1, etc in chordTabs

    for ll in l:
        if '@' in ll:  # we have something like "Cm@3.2"
            ch, beat = ll.split('@', 1)
            beat = stof(beat, "Expecting a value after the @ in '%s'" % ll)
            
            if beat < 1:
                error("Beat after @ must be 1 or greater, not '%s'." % beat)
            if beat >= gbl.QperBar + 1:
                error("Beat after @ must be less than %s, not '%s'." % (gbl.QperBar + 1, beat))

            # tick offset for this chord
            beat = int((beat - 1) * quarter)
            
            # need to set p to next spot in chordTabs[].
            for p, i in enumerate(chordTabs):
                if i > beat:
                    break
        else:
            ch = ll
            if p > len(chordTabs)-1:
                error("Too many chords specified in line. Max is %s. "
                      "For more chords use @ notation or change TIME TABS."
                      % (len(chordTabs)))
            beat = chordTabs[p]

            p += 1  # for next beat

        if ch in '-/':      # handle continuation chords
            if not ctable:
                if lastChord:
                    ch = lastChord
                else:
                    error("No previous chord for '/' or '-' at line start")
            else:  # '/' other than at start just increment the beat counter
                continue

        if ctable:
            if ctable[-1].name == ch:  # skip duplicate chords
                continue

            if ctable[-1].chStart >= beat:
                error("Chord positions out of order")

        else:    # first entry
            if beat != 0:
                error("The first chord must be at beat 1, not %s." % ((beat / quarter) + 1))

        ctab = CTable()
        ctab.name = ch
        ctab.chStart = beat
        if ctable:
            ctab.lastchord = ctable[-1].name
        else:
            ctab.lastchord = lastChord

        # If the chord we just extracted has a 'z' in it then we do the
        # following ugly stuff to figure out which tracks to mute. 'ch'
        # will be a chord name or 'z' when this is done.

        if 'z' in ch:
            if ch.count('z')>1:
                error("Only one 'z' is permitted in the mute shortcut. '%s' is "
                      "illegal." % ch)

            if lastChord and 'z' in lastChord:  # strip off z stuff from prior chord
                lastChord = lastChord.split('z')[0]
                
            c, r = ch.split('z', 1)  # get chord and track list

            if not c:   # no chord specified
                c = 'z'        # dummy chord name to keep chordnotes() happy

                if r == '!':    # 'z!' is fine, mute all
                    r = 'DCAWBSRP'
                elif not r:     # 'z' is fine, mute all tracks except Drum
                    r = 'CBAWSRP'
                    
                else:
                    if lastChord:
                        c = lastChord
                        if 'z' in c:
                            c = c.split('z')[0]  # we only want the chord
                    else:
                        error("Unable to find a chord for '%s'. Try specifing the "
                              "chord name." % ch)
                        if not c:
                            error("To use this shortcut to mute individual tracks "
                                  "the chord must be included. "
                                  "Use CHORDz[DCAWBSRP] not '%s'" % ch)


            else:    # illegal CHORD + 'z' or 'z!'
                if r == '!':
                    error("'%s' is illegal. 'z!' mutes all tracks "
                          "so you can't include the chord." % ch)
                elif not r:
                    error("'%s' is illegal. This notation is a shortcut,  You must specify tracks "
                          "if you use a chord." % ch)

            ch = c   # this will be 'z' or the chord part

            # Got a chord (real name or 'z') and tracks to mute
            for v in r:
                if v == 'C':
                    ctab.chordZ = 1
                elif v == 'B':
                    ctab.bassZ = 1
                elif v == 'A':
                    ctab.arpeggioZ = 1
                elif v == 'W':
                    ctab.walkZ = 1
                elif v == 'D':
                    ctab.drumZ = 1
                elif v == 'S':
                    ctab.scaleZ = 1
                elif v == 'R':
                    ctab.ariaZ = 1
                elif v == 'P':
                    ctab.plectrumZ = 1
                else:
                    error("Unknown track '%s' for muting in '%s'" % (v, ch))

        ctab.chord = mma.MMA.chords.ChordNotes(ch)  # Derive chord notes (or mute)
        ctable.append(ctab)

    # Finished each chord in data line, test
    # to see that all chords are in range.
    if ctable[-1].chStart >= endTime:
        error("Maximum offset for chord '%s' must be less than %s, not '%s'." %
              (ctable[-1].name, endTime / quarter + 1, ctable[-1].chStart / quarter + 1))

    # Fix up some pointers.

    for i, v in enumerate(ctable[:-1]):  # set end range for each chord
        ctable[i].chEnd = ctable[i + 1].chStart

    ctable[-1].chEnd = endTime      # set end of range for last chord
    lastChord = ctable[-1].name     # remember chord at end of this bar for next
    
    return ctable




# xtra.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Handle -x options.

"""

from common import *
from . import gbl
import macro
from mma.MMA.lyric import lyric

def checkChords(clist):
    """ Take a list of chords passed on the command line and check them
        for validity.
    """

    gbl.ignoreBadChords=True
    okaylist = []
    for b in clist:
        try:
            mma.MMA.chords.ChordNotes(b)
            okaylist.append(b)
        except:
            continue
    print ("VALID: %s" % ', '.join(okaylist))
    gbl.ignoreBadChords = False

def checkFile(l):
    """ Check a MMA input file an verify chords are valid. """

    # one could read a RC file. Insert mma.MMA.paths.readRC()
    # here, but it's probably a silly thing to do since we
    # really aren't attempting a proper parse ...

    infile = mma.MMA.file.locFile(l, None)  # filename to check
    if not infile:
        error("Input file '%s' not found." % l)
    inpath = mma.MMA.file.ReadFile(infile)
    validChords = []

    while 1:
        curline = inpath.read()
        
        if curline == None:
            break

        # If we want to support macros, begin/end, etc. insert it all
        # here. Caution: it's a deep rabbit hole, best avoided

        # check the chords in the line. All lines must start with a line #
        # if not a line # we just skip to next line.
        if curline[0].isdigit():   # curline is a list 
            l = curline[1:]
            ##  A bar can have an optional repeat count. This must
            ##  be at the end of bar in the form '* xx'. Just strip it
            if len(l) > 1 and l[-2] == '*':
                l = l[:-2]

            # join into a string
            l = ' '.join(l)
            
            # extract/discard RIFFs. Note: Malformed riffs are NOT extracted
            # in real mma code the riff/lyric can appear anywhere ... this
            # so-called parsing code is much more simplistic. Here we cut the
            # line off at the first { or [.
            l  = l.partition('{')[0].partition('[')[0]
                
            l = l.split()  
            
            for c in l:
                if c in (None, '', 'z', '/', '/!'):
                    continue

                if c in validChords:
                    continue
                
                else:
                    try:
                        gbl.ignoreBadChords = True
                        mma.MMA.chords.ChordNotes(c)
                        validChords.append(c)
                        gbl.ignoreBadChords = False
                    except:  # mma printed the error message so keep trucking
                        continue

    # summarize
    print("Valid chords: %s" % ', '.join(sorted(validChords)))
    sys.exit(0)
        
def xoption(opt, args):
    """ Xtra, seldom used, options """
    
    opt = opt.upper()

    if opt == 'NOCREDIT':
        gbl.noCredit = True
        return
        
    elif opt == 'CHORDS':
        # check a list of chords on the cmd line for validity
        if not args:
            mma.MMA.options.usage()
        checkChords(args)
        sys.exit(0)   # don't return ... all we can do is the chords

    elif opt == 'CHECKFILE':
        # check a input file for valid chords
        if len(args) != 1:
            error("-xCheckFile: Exactly one filename required. Use '-xCheckFile <FILENAME>'.")
        checkFile(args[0])
        
    else:
        error("'%s' in an unknown -x option" % opt)
             



# tempo.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

from . import gbl
from mma.MMA.common import *
from mma.MMA.midiM import packBytes, byte3ToInt
from mma.MMA.timesig import timeSig
from struct import unpack
from mma.MMA.parseCL import setChordTabs
import mma.MMA.debug
import mma.MMA.midi

######################################
# Tempo/timing

timeTable = { 
    # duple times
    '2/2': (4,    (1, 3)), 
    '2/4': (2,    (1, 2)),
    '6/4': (6,    (1, 4)),
    '6/8': (6,    (1, 4)),   # need this way to keep compatible with 'stdpats68'

    # triple
    '3/2': (6,    (1, 3, 5)),
    '3/4': (3,    (1, 2, 3)),
    '3/8': (1.5,  (1, 1.5, 2)),
    '9/8': (4.5,  (1, 2.5, 4)),
    
    # quadruple
    '4/4': (4,    (1, 2, 3, 4) ),
    '12/8':(6,    (1, 2.5, 4, 5.5)),

    # quintuple
    '5/4': (5,    (1, 2, 3, 4, 5) ),
    '5/8': (2.5,  (1, 1.5, 2, 2.5, 3 ) ), 

    # septuple
    '7/4': (7,    (1, 2, 3, 4, 5, 6, 7) )
 }

def setTime(ln):
    """ Set the 'time' value. This is NOT a time sig, it
        is the number of quarters/beat.

        We do restrict the time setting to the range of 1..12.
        No particular reason, but we do need some limit? Certainly
        it has to be greater than 0.
    """

    tabList = []
    defaultTabs = (1,2,3,4,5,6,7,8,9,10,11,12)
    sigSet = False

    ln, options = opt2pair(ln, toupper=True)

    for cmd, opt in options:
        if cmd == 'TABS':
            tabList = [ stof(i) for i in opt.split(',')]
            if tabList != sorted(tabList):
                error("Time: Tabs must be in order, not %s." % tabList)
            if len(tabList) != len(set(tabList)):
                error("Time: Tabs must not contain duplicates, %s" % tabList)
            if tabList[0] != 1:
                error("Time: First tab must be 1, not %s." % tabList[0])
        else:
            error("TIME: Unknown option '%s'." % cmd)
        
    if len(ln) != 1:
        error("Time: Too many options. Use TIME BperBar [Tabs=xx].")

    # is this a timesig?

    n = ln[0]
    if n.upper() == 'COMMON':
        n = '4/4'
    elif n.upper() == 'CUT':
        n = '2/2'

    if n in timeTable:
        i = timeTable[n]
        timeSig.setSig([n])
        n = i[0]
        if not tabList:
            tabList = i[1]
        sigSet = True
    else:
        if '/' in n:  # unknown time sig
            error("Time: Unknown timesignature '%s'. You may need to set TIME and TIMESIG separately." % n) 
        n = stof(n)

        if n < 1 or n > 12:
            error("Time: Value must be 1..12.")

    # If no change, just ignore this.

    origTime = gbl.QperBar
    if origTime != n:
        gbl.QperBar = n
        gbl.barLen =  int(gbl.QperBar * gbl.BperQ)

        # Time changes zap all predfined sequences

        for a in gbl.tnames.values():
            if a.riff:
                warning("Time: Change from %s to %s deleting %s riffs." %
                        (origTime, n, a.name))
                a.riff = []
            a.clearSequence()

    if not tabList:
        tabList = (1,2,3,4,5,6,7,8,9,10,11,12)[:int(gbl.QperBar)]

    # need to do this after setting time.
    if (tabList[-1]-1) * gbl.BperQ >= gbl.barLen:
        error("Time: Last tab must be < %s, not '%s'." % 
              (float(gbl.barLen/gbl.BperQ)+1, tabList[-1]))

    setChordTabs(tabList)

    if mma.MMA.debug.debug:
        if sigSet:
            sig =  "TimeSig %s " % timeSig.getAscii()
        else:
            sig = ''
        dPrint ("Time: Time %s %sTabs=%s." % 
               (gbl.QperBar, sig,  ','.join([str(x) for x in tabList])))
            
def tempo(ln):
    """ Set tempo.

        Note: All tempo stuff is inserted into the meta track.
    """

    if not ln or len(ln) > 2:
        error("Tempo: Use [*,+,-]BperM [BARS]")

    # Get new value.

    a = ln[0][0]
    if a in "+-*":
        v = stof(ln[0][1:], "Tempo: Expecting value for rate adjustment, not '%s'" % ln[0])
        if a == '-':
            v = gbl.tempo - v
        elif a == '+':
            v += gbl.tempo
        elif a == '*':
            v *= gbl.tempo

    else:
        v = stof(ln[0], "Tempo: Expecting rate, not '%s'" % ln[0])

    if v <= 1:
        error("Tempo: Value must be greater than 1.")

    # is this immediate or over time?

    if len(ln) == 1:
        gbl.tempo = int(v)

        gbl.mtrks[0].addTempo(gbl.tickOffset, gbl.tempo)

        if mma.MMA.debug.debug:
            dPrint("Tempo: Set to %s" % gbl.tempo)

    else:              # Do a tempo change over bar count
        bars = ln[1]

        bars = stof(bars, "Tempo: Beat count expecting value, not %s" % bars)
        numbeats = int(bars * gbl.QperBar)

        if numbeats < 1:
            error("Tempo: Beat count must be greater than 1")

        # Vary the rate in the meta track

        tincr = (v - gbl.tempo) / float(numbeats)    # incr per beat
        bstart = gbl.tickOffset            # start
        boff = 0
        tempo = gbl.tempo

        for n in range(numbeats):
            tempo += tincr
            if tempo:
                gbl.mtrks[0].addTempo(bstart + boff, int(tempo))
            boff += gbl.BperQ

        if tempo != v:
            gbl.mtrks[0].addTempo(bstart + boff, int(v))

        gbl.tempo = int(v)

        if mma.MMA.debug.debug:
            dPrint("Tempo: Set future value to %s over %s beats" % 
                (int(tempo), numbeats))

    if gbl.tempo <= 0:
        error("Tempo: Setting must be greater than 0.")


def beatAdjust(ln):
    """ Delete or insert some beats into the sequence.

        This just adjusts the current song position. Nothing is
        lost or added to the actual file.
    """

    if len(ln) != 1:
        error("BeatAdjust: Expecting single value.")

    adj = stof(ln[0], "BeatAdjust: Expecting a value (not %s)." % ln[0])

    gbl.tickOffset += int(adj * gbl.BperQ)

    gbl.totTime += adj / gbl.tempo   # adjust total time

    if mma.MMA.debug.debug:
        dPrint("BeatAdjust: inserted %s at bar %s." % (adj, gbl.barNum + 1))


def cut(ln):
    """ Insert a all-note-off into ALL tracks. """

    if not len(ln):
        ln = ['0']

    if len(ln) != 1:
        error("Cut: Expecting single offset or empty argument.")

    """ Loop though all the tracks. Note that trackCut() checks
        to make sure that there is a need to insert in specified track.
        In this loop we create a list of channels as we loop though
        all the tracks, skipping over any duplicate channels or
        tracks with no channel assigned.
    """

    l = []
    for t in sorted(gbl.tnames.keys()):
        c = gbl.tnames[t].channel
        if not c or c in l:
            continue
        l.append(c)
        trackCut(t, ln)


def trackCut(name, ln):
    """ Insert a ALL NOTES OFF at the given offset. """

    if not len(ln):
        ln = ['0']

    if len(ln) != 1:
        error("Cut %s: Offset missing." % name)

    offset = stof(ln[0], "Cut %s: Expecting value, (not '%s') for offest." % (name, ln[0]))

    if offset < -gbl.QperBar or offset > gbl.QperBar:
        warning("Cut %s: %s is a large beat offset" % (name, offset))

    moff = int(gbl.tickOffset + (gbl.BperQ * offset))

    if moff < 0:
        error("Cut %s: Offset comes before start of track." % name)

    # Insert allnoteoff directly in track. This skips the normal
    # queueing in pats because it would never take if at the end
    # of a track.

    m = gbl.tnames[name].channel
    if m and len(gbl.mtrks[m].miditrk) > 1:
        gbl.mtrks[m].addNoteOff(moff)

        if mma.MMA.debug.debug:
            dPrint("Cut %s: Beat %s, Bar %s" % (name, offset, gbl.barNum + 1))


def fermata(ln):
    """ Apply a fermata timing to the specified beat. """

    if len(ln) != 3:
        error("Fermata: use 'offset' 'duration' 'adjustment'")

    offset = stof(ln[0], "Fermata: Expecting a value (not '%s') for offset" % ln[0])

    if offset < -gbl.QperBar or offset > gbl.QperBar:
        warning("Fermata: %s is a large beat offset" % offset)

    dur = stof(ln[1], "Fermata: Expecting a value (not '%s') for duration" % ln[1])

    if dur <= 0:
        error("Fermata: duration must be greater than 0")

    if dur > gbl.QperBar:
        warning("Fermata: %s is a large duration." % dur)

    adj = stof(ln[2], "Fermata: expecting a value (not '%s') for adjustment." % ln[2])

    if adj < 100:
        warning("Fermata: Adjustment less than 100 is shortening beat value.")

    if adj == 100:
        error("Fermata: using value of 100 makes no difference, must be an error.")

    moff = int(gbl.tickOffset + (gbl.BperQ * offset))

    if moff < 0:
        error("Fermata: Offset comes before track start.")

    if offset >= 0:
        warning("Fermata: Better results when placed after event (negative offset).")

    fermataDuration = int(gbl.BperQ * dur)  # Duration in ticks
    mend = moff + fermataDuration

    # This next section is needed to figure out the start tempo (which
    # is not always gbl.tempo!), the needed tempo at the end of the fermata
    # section (again, not always gbl.tempo!) and to delete any tempo changes
    # in the section. All this due to the effects of tempo changes over a
    # a set of bars.

    # Extract ALL tempo changes from the meta track and save in a list.
    # Also, delete any tempo changes found in the miditrack in our range.

    tempos = []
    tcmd = packBytes((0xff, 0x51, 0x03))   # midi TEMPO
    mt = gbl.mtrks[0].miditrk  # The meta track

    # 1st we copy existing tempos into a new list
    for t in mt:
        for ev in mt[t]:
            if ev[0:3] == tcmd:
                tempos.append((t, ev[3:]))  # save the 24 bit value

    # now we delete any tempos in the fermata range. This avoids broken loops
    for t, ev in tempos:
        if t >= moff and t <= mend:
            gbl.mtrks[0].delDup(t, tcmd)

    # find last tempo before fermata bock and last tempo in block
    tempos.sort()
    oldTempo = gbl.tempo
    newTempo = gbl.tempo

    for f, t in tempos:   # f==offsets, t==encoded tempos
        if f <= moff:
            oldTempo = 60000000 // byte3ToInt(t)
        if f >= moff and f <= mend:
            newTempo = 60000000 // byte3ToInt(t)

    gbl.mtrks[0].addTempo(moff, int(oldTempo / (adj / 100)))
    gbl.mtrks[0].addTempo(mend, newTempo)

    # Move selected events in the effected area to it's start/end.
    #   To start -> note on, program change
    #   to end   -> note off
    # Done only if the fermata comes after desired location ...
    # otherwise there aren't any events to zap!

    if offset < 0:
        for n, tr in gbl.mtrks.items():  # do each track
            if n <= 0:
                continue        # skip meta track

            trk = gbl.mtrks[n].miditrk
            startEvents = []
            endEvents = []

            for f in sorted(trk):          # all in this track (sort keeps orig order)
                if f > moff and f < mend:  # well, only in the fermata range
                    remain = []
                    for ev in trk[f]:      # all events in the offset
                        if not ev:         # skip empty events
                            continue
                        # Get event type. The [0:1] is needed to
                        # maintain this as a byte() for python3, otherwise
                        # py3 will think it's an int() and barfs
                        evtype = unpack('B', ev[0:1])[0] >> 4
                        if evtype == 0x9:         # note event
                            if ev[2] == 0:    # off to end
                                endEvents.append(ev)
                            else:
                                startEvents.append(ev)  # on to start
                            continue
                        if evtype == 0xb or evtype == 0xc:  # program/controller change
                            startEvents.append(ev)          # all to start (??)
                            continue
                        remain.append(ev)
                    trk[f] = remain        # remaining events for this offset

            if startEvents:
                if moff in trk:
                    trk[moff].extend(startEvents)
                else:
                    trk[moff] = startEvents

            if endEvents:
                if mend in trk:
                    trk[mend] = endEvents + trk[mend]
                else:
                    trk[mend] = endEvents

    if mma.MMA.debug.debug:
        dPrint("Fermata: Beat %s, Duration %s, Change %s, Bar %s" % 
              (offset, dur, adj, gbl.barNum + 1))
        if offset < 0:
            dPrint("         NoteOn Events moved in tick range %s to %s" 
                  % (moff + 1, mend - 1))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mma",
    version="1.1.20.02",
    author="Hugo Neves de Carvalho",
    author_email="hugonvsc@gmail.com",
    description="MMA - Music Midi Accompaniment available as a pip module",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hugonxc/mma_pip",
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points ={ 
        'console_scripts': [ 
            'mma = mma.mma:cli'
        ] 
    }, 
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


# player.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import time
import subprocess
import re

from mma.MMA.common import *
from . import gbl
import mma.MMA.file
import mma.MMA.debug

# Initialize the default midi player.
if gbl.platform == 'Windows':
    midiPlayer = ['']   # must be a list!
elif gbl.platform == 'Darwin':
    midiPlayer = ['open']   # must be a list!
else:
    midiPlayer = ["aplaymidi"] # Must be a list!

# We run in background in windows, foreground in linux
if gbl.platform == 'Windows':
    inBackGround = 1  # by default we run in foreground
else:
    inBackGround = 0

waitTime = 5  # default time to wait after forking in background


def setMidiPlayer(ln):
    """ Set the MIDI file player (used with -P and -V). """

    global midiPlayer, waitTime, inBackGround

    if not ln:
        ln = ['']

    n = []
    for l in ln:   # parse out optional args
        if '=' in l and l[0].isalpha():
            a, b = l.upper().split('=', 1)
            if a == 'DELAY':
                b = stof(b, "SetMidiPlayer: Delay must be value, not '%s'." % b)
                waitTime = b

            elif a == "BACKGROUND":
                if b in ('1', 'YES'):
                    inBackGround = 1
                elif b in ('0', 'NO'):
                    inBackGround = 0
                else:
                    error("SetMidiPlayer: Background must be 'yes'"
                          "or 'no', not '%s'." % b)

            else:
                error("SetMidiPlayer: unknown option '%s'." % a)

        else:
            n.append(mma.MMA.file.fixfname(l))

    if not n:
        n = ['']
    midiPlayer = n

    if mma.MMA.debug.debug:
        dPrint("MidiPlayer set to '%s' Background=%s Delay=%s." %
            (' '.join(midiPlayer), inBackGround, waitTime))


def playMidi(file):
    """ Play a midi file. """

    pl = midiPlayer[0]
    opts = midiPlayer[1:]

    if not pl and gbl.platform != "Windows":
        error("No MIDI file player defined, temp files will be deleted.")

    if not pl:
        m = "default windows MIDI player"
    else:
        m = pl
    print("Playing MIDI '%s' with %s." % (file, m))

    if gbl.platform == "Windows":
        sh = True
    else:
        sh = False

    cmd = [pl]
    if opts:
    #    cmd.append(' '.join(opts))  # puts all opts into one string, not right??
        cmd.extend(opts)
    cmd.append(file)

    t = time.time()

    # fork our player.
    try:
        pid = subprocess.Popen(cmd, shell=sh)
    except OSError as e:
        dPrint(e)
        msg = "MidiPlayer fork error."
        if re.search("[\'\"]", ''.join(cmd)):
            msg += " Using quotes in the MidiPlayer name/opts might be your problem."
        error(msg)

    if inBackGround:    # if the background option set, do a sleep
        print("Play in progress ... file will be deleted.")
        time.sleep(waitTime)

    else:   # foreground player ... wait for process to finish
        pid.wait()
        print("Play complete (%.2f min), MIDI file has been deleted." 
            % ((time.time() - t) / 60))



# exits.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module contains cleanup code to be called on exit of mma.MMA.

"""

import atexit
import os
import mma.MMA.common

files = []


def cleanup():
    """ This cleanup routine will delete registered files. Currently this
        includes:

          files ... created by groove preview and autoplay funcs
    """

    for f in files:
        try:
            os.remove(f)
        except:
            pass

    mma.MMA.common.cleanPrintBuffer()

atexit.register(cleanup)



# patArpeggio.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import random

import mma.MMA.notelen
import mma.MMA.harmony

from . import gbl
from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup


class Arpeggio(PC):
    """ Pattern class for an arpeggio track. """

    vtype = 'ARPEGGIO'
    arpOffset = -1
    arpDirection = 1
    lastRange = 999
    lastDirection = 999

    def getPgroup(self, ev):
        """ Get group for apreggio pattern.

            Fields - start, length, volume
        """

        a = Pgroup()
        if len(ev) != 3:
            error("There must be exactly 3 items in each group "
                  "for apreggio define, not '%s'" % ' '.join(ev))

        a.offset = self.setBarOffset(ev[0])
        a.duration = mma.MMA.notelen.getNoteLen(ev[1])
        a.vol = stoi(ev[2], "Type error in Arpeggio definition")

        return a

    def restart(self):
        self.ssvoice = -1
        self.arpOffset = -1
        self.arpDirection = 1
        lastRange = 999
        lastDirection = 999

    def trackBar(self, pattern, ctable):
        """ Do a arpeggio bar.

        Called from self.bar()

        """

        sc = self.seq

        direct = self.direction[sc]
        if direct != self.lastDirection:
            self.arpOffset = -1
            self.arpDirection = 1
            self.lastDirection = direct

        crange = self.chordRange[sc]
        if crange != self.lastRange:
            self.arpOffset = -1
            self.arpDirection = 1
            self.lastRange = crange

        for p in pattern:
            tb = self.getChordInPos(p.offset, ctable)

            if tb.arpeggioZ:
                continue

            if direct == 'DOWN':
                self.arpDirection = -1

            if self.chordLimit:
                tb.chord.limit(self.chordLimit)

            if self.compress[sc]:
                tb.chord.compress()

            if self.invert[sc]:
                tb.chord.invert(random.randrange(self.invert[sc][0], self.invert[sc][1]+1))

            # This should be optimized, it recreates the chord for every pattern.
            # Problem is that one would need to check all the LIMIT, COMPRESS, etc
            # settings each for each bar as well, so it's probably just as easy to
            # leave it as is. Besides, this works.

            octave = 0  # octave adjuster
            ourChord = []
            
            # for the range, stack up chords
            for l in range(int(crange)):
                ourChord.extend([x+octave for x in tb.chord.noteList])
                octave += 12

            # fractional range, add partial chord
            fr = crange - int(crange)
            if fr:      # for fractional  lengths
                fr = int(tb.chord.noteListLen * fr)
                if fr < 2:   # important, min of 2 notes in arp.
                    fr = 2
                ourChord.extend([x+octave for x in tb.chord.noteList[:fr]])

            if direct == 'BOTH':
                if self.arpOffset < 0:
                    self.arpOffset = 1
                    self.arpDirection = 1
                elif self.arpOffset >= len(ourChord):
                    self.arpOffset = len(ourChord) - 2
                    self.arpDirection = -1

            elif direct == 'UP':
                if self.arpOffset >= len(ourChord) or self.arpOffset < 0:
                    self.arpOffset = 0
                    self.arpDirection = 1

            elif direct == 'DOWN':
                if self.arpOffset < 0 or self.arpOffset >= len(ourChord):
                    self.arpOffset = len(ourChord) - 1
                    self.arpDirection = -1

            if direct == 'RANDOM':
                note = random.choice(ourChord)
            else:
                note = ourChord[self.arpOffset]

            self.arpOffset += self.arpDirection

            if not self.harmonyOnly[sc]:
                notelist = [(note, p.vol)]
            else:
                notelist = []

            if self.harmony[sc]:
                h = mma.MMA.harmony.harmonize(self.harmony[sc], note, ourChord)
                vol = p.vol * self.harmonyVolume[sc]
                harmlist = list(zip(h, [vol] * len(h)))
            else:
                harmlist = []

            offset = p.offset
            if self.ornaments['type']:
                offset = mma.MMA.ornament.doOrnament(self, notelist,
                               self.getChordInPos(offset, ctable).chord.scaleList, p)
                notelist = []

            self.sendChord(notelist + harmlist, p.duration, offset)

            tb.chord.reset()    # important, other tracks chord object



# patScale.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import random

import mma.MMA.harmony
import mma.MMA.notelen
import mma.MMA.ornament
from mma.MMA.pat  import PC, Pgroup

from . import gbl
from mma.MMA.common import *
import mma.MMA.debug

import copy


class Scale(PC):
    """ Pattern class for a Scale track. """

    vtype = 'SCALE'

    lastNote = -1
    lastChord = None
    lastStype = None
    lastDirect = 1
    lastRange = 0
    sOffset = 0
    notes = None
    dirfact = 1

    def __init__(self, ln):
        
        PC.__init__(self, ln)
        
    def saveGroove(self, gname):
        """ Save special/local variables for groove. """

        PC.saveGroove(self, gname)  # create storage. Do this 1st.

    def restoreGroove(self, gname):
        """ Restore special/local/variables for groove. """

        PC.restoreGroove(self, gname)

    def getPgroup(self, ev):
        """ Get group for scale patterns.

            Fields - start, length, volume
        """

        if len(ev) != 3:
            error("There must be at exactly 3 items in each group "
                  "in a Scale definition, not <%s>." % ' '.join(ev))

        a = Pgroup()

        a.offset = self.setBarOffset(ev[0])
        a.duration = mma.MMA.notelen.getNoteLen(ev[1])
        a.vol = stoi(ev[2], "Type error in Scale definition")

        return a

    def setScaletype(self, ln):
        """ Set scale type. """

        ln = lnExpand(ln, "%s ScaleType" % self.name)
        tmp = []

        for n in ln:
            n = n.upper()
            if not n in ( 'CHROMATIC', 'SCALE', 'AUTO', 'CHORD'):
                error("%s Scaletype: Only Chromatic, Scale, Auto and "
                      "Chord are valid, not '%s'" % (self.name, n))

            tmp.append(n)

        self.scaleType = seqBump(tmp)

        if mma.MMA.debug.debug:
            dPrint("Set %s ScaleType to: %s" %
                  (self.name, ' '.join(self.scaleType)))
        
    def restart(self):
        self.ssvoice = -1
        self.lastNote = -1
        self.lastChord = None
        self.lastStype = None
        self.lastDirect = 1
        self.lastRange = 0
        self.sOffset = 0
        self.notes = None
        self.dirfact = 1

        
    def trackBar(self, pattern, ctable):
        """ Do a scale bar.

            Called from self.bar()
        """

        sc = self.seq
        direct = self.direction[sc]
        
        # If the range or direction has changed, we just start
        # with a new scale.

        t = self.chordRange[sc]
        if t != self.lastRange:
            self.lastRange = t
            self.lastChord = None

        if self.lastDirect != direct:
            self.lastDirect = direct
            self.lastChord = None

        for p in pattern:

            tb = self.getChordInPos(p.offset, ctable)

            if tb.scaleZ:
                continue

            thisChord = tb.chord.tonic + tb.chord.chordType
            stype = self.scaleType[sc]

            if thisChord != self.lastChord or stype != self.lastStype:
                self.lastChord = thisChord
                self.lastStype = stype

                if stype == 'CHROMATIC':
                    notelist = [ tb.chord.rootNote + x for x in range(0,12)]
                
                elif stype == 'CHORD':
                    notelist = tb.chord.noteList[:]

                else:
                    notelist = list(tb.chord.scaleList)
                
                """ Get the current scale and append enuf copies
                together for RANGE setting. If Range happens
                to be 0 or 1 we end up with a single copy.
                """

                ln=self.chordRange[sc]    # RANGE 1...x (def. == 1)

                o=0
                self.notes = []

                while ln >= 1:
                    for a in notelist:
                        self.notes.append(a+o)
                    o+=12
                    ln-=1

                if ln>0 and ln<1:  # for fractional scale lengths
                    ln = int(len(notelist) * ln)
                    if ln < 2:   # important, must be at least 2 notes in a scale
                        ln=2
                    for a in notelist[:ln]:
                        self.notes.append(a+o)

                if direct == 'DOWN':
                    self.dirfact = -1
                    if self.lastNote == -1:
                        self.sOffset = len(self.notes)-1
                else:
                    self.sOffset = 0
                    self.dirfact = 1

                if self.lastNote > -1:
                    if self.lastNote in self.notes:
                        self.sOffset = self.notes.index(self.lastNote)

                    else:
                        self.sOffset=len(self.notes)-1
                        for     i, a in enumerate(self.notes):
                            if a>self.lastNote:
                                self.sOffset = i
                                break


            # Keep offset into note list in range

            # only > end of list if BOTH or UP

            if self.sOffset >= len(self.notes):
                if direct == 'BOTH':
                    self.dirfact = -1
                    self.sOffset = len(self.notes)-2
                else:        ## UP
                    self.sOffset = 0

            # only < start of list if DOWN or BOTH

            elif self.sOffset < 0:
                if direct == 'BOTH':
                    self.dirfact = 1
                    self.sOffset = 1
                else:        ## DOWN
                    self.sOffset = len(self.notes)-1

            if direct == 'RANDOM':
                note = random.choice(self.notes)
            else:
                note = self.notes[self.sOffset]
                self.sOffset += self.dirfact

            self.lastNote = note

            duration = p.duration
            vol = p.vol
            offset = p.offset

            if not self.harmonyOnly[sc]:
                nlist = [(note, p.vol)]
            else:
                nlist = []

            if self.harmony[sc]:
                ch = self.getChordInPos(p.offset, ctable).chord.noteList
                h = mma.MMA.harmony.harmonize(self.harmony[sc], note, ch)
                harmlist =  list(zip(h, [p.vol * self.harmonyVolume[sc]] * len(h)))
            else:
                harmlist = []

            offset = p.offset
            if self.ornaments['type']:
                offset = mma.MMA.ornament.doOrnament(self, nlist,
                                        self.getChordInPos(offset, ctable).chord.scaleList, p)
                nlist = []

            self.sendChord(nlist+harmlist, p.duration, offset)

